"""
AWS utilities
"""
import os
import numpy as np
from .. import utils

def create_s3_index(path, output_file="index.html", extra_html="", upload=True, as_table=True, with_dja_css=True, show_images=True):
    """
    Create HTML listing of an S3 path
    """
    import time
    from .. import utils as grizli_utils
    
    url = path.replace('s3://', 'https://s3.amazonaws.com/')

    lsfile = output_file.replace('.html','.ls')

    # Sort by filename
    sort_str = 'sort -k 4'

    os.system(f'aws s3 ls {path} | {sort_str} |grep -v PRE |grep -v {output_file} > {lsfile}')
    
    now = time.ctime()
    html=f"<h3>{now}</h3>\n"

    html += extra_html

    html += "\n<pre>\n"

    lines=open(lsfile).readlines()

    rows = []

    for line in lines:
        lsp = line.split()
        if len(lsp) == 4:
            html += "{0} {1} {2:>10} <a href={4}{3} > {3} </a>\n".format(*lsp, url)
            rows.append(lsp)
            
    html += '</pre>\n'
    
    if as_table:
        names = ['date', 'time', 'size', 'file']
        tab = grizli_utils.GTable(rows=rows, names=names)
        size = np.asarray(tab['size'],dtype=float)
        tab['size'] = [f'{s/1.e6:.2f}' for s in size]
        tab['date'] = ['{date} {time}'.format(**row) for row in tab]
        tab.remove_column('time')
        
        tab['size'] = np.asarray(tab['size'],dtype=float)
        
        dl = []
        for file in tab['file']:
            _uf = f'{url}{file}'
            if (file.split('.')[-1] in ['png','jpg']) & show_images:
                dl.append(f'<a href="{_uf}" /> <img src="{_uf}" height=200px /> </a> ')
            else:
                dl.append(f'<a href="{_uf}" > {_uf} </a>')
        
        tab['download'] = dl
        
        tab.write_sortable_html(output_file,
                                use_json=False,
                                localhost=False,
                                max_lines=1e6,
                                with_dja_css=with_dja_css,
                                toggle=False,
                                #filter_columns=['size'],
                                buttons=['csv'],
                                )
        
        tab.write(output_file.replace('.html', '.csv'), overwrite=True)
        
        if extra_html:
            with open(output_file) as fp:
                lines = fp.readlines()
            
            for i, line in enumerate(lines):
                if '<table' in line:
                    lines.insert(i, extra_html)
                    break
            
            with open(output_file,'w') as fp:
                fp.writelines(lines)
                
    else:

        fp = open(output_file,'w')
        fp.write(html)
        fp.close()
    
    if upload:
        os.system(f'aws s3 cp {output_file} {path} --acl public-read')
        os.system(f'aws s3 cp {output_file.replace(".html", ".csv")} {path} ' + 
                  ' --acl public-read')
        print(path.replace('s3://', 'https://s3.amazonaws.com/') + output_file)
    else:
        print(output_file)


def generate_program_codes(write=True):
    """
    Generate program codes tables from DB query
    """
    from . import db
    
    hst = db.SQL("""select proposal_id, max(proposal_pi) as PI, count(proposal_id), max(substr("dataURL", 19, 3)) as prog_code
    from assoc_table
    where instrument_name in ('WFC3/IR','ACS/WFC','WFC3/UVIS')
    group by proposal_id
    order by CAST(proposal_id as INT)
    """)

    jw = db.SQL("""select proposal_id, min(proposal_pi) as PI, count(proposal_id)
    from assoc_table
    where instrument_name in ('NIRISS','NIRCAM','MIRI')
    group by proposal_id
    order by CAST(proposal_id as INT)
    """)
    
    jw['pi'][jw['proposal_id'] == '2514'] = 'Williams, Christina'
    
    if write:
        hst.write('/tmp/hst_program_codes.csv', overwrite=True)
        jw.write('/tmp/jwst_program_codes.csv', overwrite=True)
    
    return hst, jw

def program_ids_in_mosaic(wcs_file='https://s3.amazonaws.com/grizli-v2/JwstMosaics/v7/gds-grizli-v7.0-f850lp_wcs.csv', progs_list=None, return_raw_list=False):
    """
    Generate a summary of observing programs that contribute to a particular grizli mosaic
    
    Parameters
    ----------
    wcs_file : str
        Filename or URL of a wcs summary file produced by `~grizli.aws.visit_processor.cutout_mosaic`
    
    progs_list : list
        List of parsed programs from ``return_raw_list=True``, e.g., for concatenating
        multiple lists into a single table
    
    return_raw_list : bool
        Just return the list of parsed programs for each exposure
    
    Returns
    -------
    tab : list, `~astropy.table.Table`
        List of parsed programs or a full table with program summary
    
    """
    # Read summary tables
    #--------------------
    data_path = os.path.join(os.path.dirname(__file__), '../data')
    
    # HST
    #-------------
    hst_file = os.path.join(data_path, 'hst_program_codes.csv')
    if not os.path.exists(hst_file):
        hst_file = 'https://dawn-cph.github.io/dja/data/hst_program_codes.csv'
        
    hst = utils.read_catalog(hst_file)
    hst_programs = {}
    hst_pi = {}
    for row in hst:
        hst_programs[row['prog_code']] = row['proposal_id']
        hst_pi[str(row['proposal_id'])] = row['pi']
    
    # JWST
    #--------------
    jw_file = os.path.join(data_path, 'jwst_program_codes.csv')
    if not os.path.exists(jw_file):
        jw_file = 'https://dawn-cph.github.io/dja/data/jwst_program_codes.csv'
    
    jw = utils.read_catalog(jw_file)
    jwst_pi = {}
    for row in jw:
        jwst_pi[str(row['proposal_id'])] = row['pi']

    # Loop through file names and calculate program IDs
    #--------------------------------------------------
    if progs_list is None:
        
        # grizli WCS summary file, remote or local
        #------------------------------------------
        exp = utils.read_catalog(wcs_file)
        
        progs_list = []
        for file in np.unique(exp['file']):
            if file.startswith('jw'):
                progs_list.append(f'JWST-{file[3:7]}')
            else:
                progs_list.append(f'HST-{hst_programs[file[1:4]]}')
    
    if return_raw_list:
        return progs_list
        
    # Generate output table
    #----------------------
    un = utils.Unique(progs_list, verbose=False)
    
    url = 'https://www.stsci.edu/cgi-bin/get-proposal-info?id={0}&observatory={1}'
    rows = []
    for v in un.values:
        obs, prog = v.split('-')
        pi = hst_pi[prog] if obs == 'HST' else jwst_pi[prog]
        rows.append([int(prog), obs, un[v].sum(), pi, url.format(prog, obs)])
    
    tab = utils.GTable(rows=rows,
                       names=['program_id', 'observatory', 'count', 'pi', 'url'])
    
    so = np.argsort(tab['program_id'])
    tab = tab[so]
    tab.meta['wcs_file'] = wcs_file
    
    return tab


"""
Reprocessing scripts for variable WFC3/IR backgrounds
"""
from .. import utils


def reprocess_wfc3ir(parallel=False, cpu_count=0, clean_dark_refs=True):
    """
    Run reprocessing script to flatten IR backgrounds

    """
    if parallel:
        # Might require light backend to avoid segfaults in multiprocessing
        import matplotlib as mpl
        mpl.rcParams['backend'] = 'agg'

    import glob
    import os
    import astropy.io.fits as pyfits

    # https://github.com/gbrammer/wfc3
    try:
        from reprocess_wfc3 import reprocess_wfc3
    except:
        try:
            from mywfc3 import reprocess_wfc3
        except:
            print("""
    Couldn\'t `import reprocess_wfc3`.
    Get it from https://github.com/gbrammer/reprocess_wfc3 """)
            return False

    # Fetch calibs in serial
    print('\ngrizli.pipeline.reprocess: Fetch calibrations...\n')
    files = glob.glob('*raw.fits')
    for file in files:
        reprocess_wfc3.fetch_calibs(file, ftpdir='https://hst-crds.stsci.edu/unchecked_get/references/hst/', verbose=False)

    # Make ramp diagnostic images
    if parallel:
        files = glob.glob('*raw.fits')
        files.sort()
        reprocess_wfc3.show_ramps_parallel(files, cpu_count=cpu_count)

        # Reprocess all raw files
        files = glob.glob('*raw.fits')
        files.sort()
        reprocess_wfc3.reprocess_parallel(files, cpu_count=cpu_count)
    else:
        files = glob.glob('*raw.fits')
        files.sort()

        for file in files:
            if not os.path.exists(file.replace('raw.fits', 'ramp.png')):
                reprocess_wfc3.show_MultiAccum_reads(raw=file, stats_region=[[300, 700], [300, 700]])

        for file in files:
            if not os.path.exists(file.replace('raw.fits', 'flt.fits')):
                reprocess_wfc3.make_IMA_FLT(raw=file, stats_region=[[300, 700], [300, 700]])

    # Remove DRK reference files, which now proliferate
    if clean_dark_refs:
        files = glob.glob('*raw.fits')
        files.sort()
        for file in files:
            im = pyfits.open(file)
            dark_file = im[0].header['DARKFILE'].replace('iref$', '')
            dark_file = os.path.join(os.getenv('iref'), dark_file)
            im.close()
            if os.path.exists(dark_file):
                print('clean_dark: {0}'.format(dark_file))
                os.remove(dark_file)


def inspect(root='grizli', force=False):
    """
    Run the GUI inspection tool on the `ramp.png` images to flag problematic
    reads with high backgrounds and/or satellite trails.

    Click the right mouse button to flag a given read and go to the next
    object with the 'n' key.

    Type 'q' when done.

    Parameters
    ----------
    root : str
        Rootname for the output inspection file:

            >>> root = 'grizli'
            >>> file = '{0}_inspect.fits'.format(root)

    Returns
    -------
    Nothing returned.  Makes the inspection file and runs the reprocessing.


    .. note:: If the script fails puking lots of Tk-related messages, be sure
              to run this script iin a fresh python session *before* importing
              `~matplotlib`.

    """
    import os
    import glob

    import matplotlib
    matplotlib.use("TkAgg")  # This needs to be run first!

    #import mywfc3.reprocess_wfc3
    from reprocess_wfc3 import reprocess_wfc3

    import astropy.io.fits as pyfits
    import numpy as np

    files = glob.glob('*ramp.png')
    files.sort()

    # Run the GUI, 'q' to quit
    try:
        import mywfc3.inspect
        if os.path.exists('{0}_inspect.fits'.format(root)):
            if force:
                x = mywfc3.inspect.ImageClassifier(images=files,
                                           logfile='{0}_inspect'.format(root))
        else:
            x = mywfc3.inspect.ImageClassifier(images=files,
                                       logfile='{0}_inspect'.format(root))
    except:
        pass

    if not os.path.exists('{0}_inspect.fits'.format(root)):
        return True

    #im = pyfits.open('inspect_raw.info.fits')
    im = pyfits.open('{0}_inspect.fits'.format(root))
    tab = im[1].data

    fl = im['FLAGGED'].data
    is_flagged = fl.sum(axis=1) > 0

    sat_files = [file.replace('ramp.png', 'flt.fits') for file in tab['images'][is_flagged]]

    read_idx = np.arange(14, dtype=int)+1
    idx = np.arange(fl.shape[0])

    for i in idx[is_flagged]:

        pop_reads = list(read_idx[fl[i, :] > 0])

        raw = tab['images'][i].replace('_ramp.png', '_raw.fits')

        ramp_file = tab['images'][i].replace('_ramp.png', '_ramp.dat')
        #sn_pop = mywfc3.inspect.check_background_SN(ramp_file=ramp_file, show=False)
        #pop_reads = np.asarray(np.unique(np.hstack((pop_reads, sn_pop))),dtype=int)
        #pop_reads = list(pop_reads)

        flt = raw.replace('_raw', '_flt')
        if os.path.exists(flt):
            flt_im = pyfits.open(flt)
            if 'NPOP' in flt_im[0].header:
                if flt_im[0].header['NPOP'] > 0:
                    print('Skip %s' % (flt))
                    continue

        print('Process %s %s' % (raw, pop_reads))

        reprocess_wfc3.make_IMA_FLT(raw=raw, pop_reads=pop_reads, flatten_ramp=True)

    # Remove "killed"
    kill_files = [file.replace('ramp.png', 'flt.fits') for file in tab['images'][tab['kill'] > 0]]
    for file in kill_files:
        if os.path.exists(file):
            os.remove(file)

    return True


def make_masks(files=None, inspect='grizli_inspect.fits', ext=1):
    """
    Make satellite trail masks
    """
    import os
    import numpy as np
    import astropy.io.fits as pyfits
    
    try:
        from .. import utils
        from ..ds9 import DS9
    except:
        from grizli import utils
        from grizli.ds9 import DS9

    if files is None:
        insp = utils.read_catalog(inspect)
        flag = (insp['satellite'] > 0) | (insp['earth'] > 0)
        files = [f.replace('_ramp.png', '_flt.fits') for f in insp['images'][flag]]

    ds9 = DS9()
    for file in files:
        im = pyfits.open(file)
        med = np.median(im['SCI', ext].data)
        ds9.view(im['SCI', ext].data-med)
        reg_file = file.replace('_flt.fits', '.*.mask.reg').replace('_flc.fits', '.*.mask.reg').replace('_c0m.fits', '.*.mask.reg').replace('*', '{0:02d}'.format(ext))
        if os.path.exists(reg_file):
            ds9.set('regions file '+reg_file)

        x = input(file+': draw region (x to skip, q to abort): ')
        if x in ['q']:
            print('Abort.')
            # continue
            return False
        elif x in ['x']:
            print('Skip {0}.'.format(file))
            continue

        ds9.set('regions save '+reg_file)



"""
Splitting fields into tiles
"""
import os
import glob
import warnings

import numpy as np
import astropy.wcs as pywcs
import astropy.time

from collections import OrderedDict
import yaml

from .. import utils

if os.path.exists('/GrizliImaging'):
    HOME = '/GrizliImaging'
else:
    HOME = os.getcwd()

def make_all_fields():
    """
    Define tile layout
    """
    from grizli.aws.field_tiles import make_field_tiles
    
    layout = """# field ra dec dx dy
    abell370   39.9750000 -1.5752778   40 40 
    macs0416   64.0333333 -24.0730556  40 40  
    macs1149   177.4000000 22.3972222  40 40  
    abell2744   3.5875000 -30.3966667  40 40
    macs0417   64.3958333 -11.9091667  40 40 
    macs1423   215.9500000 24.0811111  40 40 
    sgas1723    260.895 34.205 20 20
    sgas1226    186.69  21.88  10 10
    dracoii    238.1983333 64.56528 20 20
    whl0137    24.355  -8.457 22 22
    abell1689   197.8765  -1.328 46 46
    cos      150.125 2.2 90 90
    j013804m2156   24.51616  -21.93046 40 40 
    j015020m1006   27.58366  -10.09244 40 40 
    j045412m0302   73.547   -3.013     40 40 
    j091840m8103  139.64290  -81.05227 40 40 
    j134200m2442  205.50969  -24.69773 40 40 
    j152256p2536  230.72345   25.59692 40 40 
    j212928m0741  322.358   -7.69209 40 40 
    gds 53.1592464 -27.7828223 46 46
    egs 214.8288    52.8067333 60 60
    gdn 189.2369029 62.2312839 46 46
    uds 34.40869 -5.16299 30 30
    smacs0723 110.83403 -73.45429 20 20
    j1235 189.025 4.948 20 20
    macsj0647 101.9482378 70.2297032 46 46
    abells1063   342.1839985 -44.5308919 46 46  
    spitzer_idf 265.0347875 68.9741119 46 46
    sextansa 152.7749473 -4.7061916 46 46
    rxcj0600 90.04 -20.145 46 46
    ulasj1342 205.555 9.462 46 46
    izw18 143.5087276 55.2404070 46 46
    q2343p12 356.60847 12.80712 46 46
    abell1703 198.76667958333334 51.821752 46 46
    sunburst 237.5295615 -78.1917929 46 46
    abell2764 5.713566 -49.24966 90 90
    abell2390 328.397307 17.709771 90 90
    gama100033 130.58659471 1.61764667 46 46
    clg-j1212p2733 183.08573496 27.57669020 46 46
    plck-g165p67 171.78782682 42.47523529 46 46
    plck-g191p62 161.16195126 33.83078874 46 46
    tn-j1338m1942 204.61036521 -19.67247443 46 46
    nep-tdf 260.6987363 65.8226460 46 46
    spiderweb 175.205 -26.489 35.4987 35.4987
    udsxl 34.29 -5.18 57.344 57.344
    """
    
    # 46 46 for tile ref 9 9
    
    tile_defs = utils.read_catalog(layout)
    
    # size: (npix // 9 * 8) * pscale/60 * 10.5 # for even number
    # size: (npix // 9 * 8) * pscale/60 * 5 # for odd number
    
    if 0:
        field = 'gdn'
        field = 'macsj0647'
        field = 'macs1423'
        field = 'abells1063'
        
        # abell2764, 2390: pscale=0.3 Euclid ERO
        npix = 2048+256
        pscale = 0.08
        ix = np.where(tile_defs['field'] == field)[0][0]
        tile_defs['rsize'] = tile_defs['dx']/2
        
        if (field in ['spiderweb']) | (1):
            npix = 4096 + 512
            
        tiles = make_field_tiles(**tile_defs[ix], tile_npix=npix, pscale=pscale, 
                                 initial_status=90, send_to_database=True)
    
    return tile_defs


def define_tiles(ra=109.3935148, dec=37.74934031, size=(24, 24), tile_size=6, overlap=0.3, field='macs0717', pixscale=0.05, theta=0):
    """
    Tile definition
    """
    
    tile_wcs = OrderedDict()
    
    size_per = tile_size-overlap
    nx = int(np.ceil(size[0]/size_per))
    ny = int(np.ceil(size[1]/size_per))
    
    sx = nx*tile_size-(nx-1)*overlap
    sy = ny*tile_size-(ny-1)*overlap
    px = int(tile_size*60/pixscale)

    # Even number of pixels
    header, parent_wcs = utils.make_wcsheader(ra=ra, dec=dec,
                                              size=(sx*60, sy*60), 
                                              pixscale=pixscale, theta=theta)
    for i in range(nx):
        x0 = int(i*px-i*(overlap/tile_size*px))
        slx = slice(x0, x0+px)
        for j in range(ny):
            y0 = int(j*px-j*(overlap/tile_size*px))
            sly = slice(y0, y0+px)
            
            slice_header = utils.get_wcs_slice_header(parent_wcs, slx, sly)
            slice_wcs = pywcs.WCS(slice_header)
            slice_wcs._header = slice_header
            #tile=(i+1)*10+(j+1)
            tile = '{0:02d}.{1:02d}'.format(i+1, j+1)
            tile_wcs[tile] = slice_wcs
    
    tile_dict = {}
    for k in tile_wcs:
        _h = utils.to_header(tile_wcs[k])
        tile_dict[k] = {'field':field}
        for key in _h:
            tile_dict[k][key] = _h[key]
    
    yml_file = '{0}_{1:02d}mas_tile_wcs.yaml'.format(field, int(pixscale*1000))
    with open(yml_file,'w') as fp:
        yaml.dump(tile_dict, stream=fp)
                                                          
    #np.save(,[tile_wcs])
    reg_file = yml_file.replace('.yaml', '.reg')
    with open(reg_file,'w') as fpr:
        fpr.write('fk5\n')
        for t in tile_wcs:
            fp = tile_wcs[t].calc_footprint()
            pstr = 'polygon('
            pstr += ','.join(['{0:.6f}'.format(i)
                            for i in fp.flatten()])
            pstr += ') # text={{{0}}}\n'.format(t)
            fpr.write(pstr)
    
    return tile_wcs


def make_field_tiles(field='cos', ra=150.125, dec=2.2, rsize=45, tile_npix=2048+256, pscale=0.08, initial_status=90, send_to_database=False, **kwargs):
    """
    Make tiles across COSMOS with a single tangent point
    """
    import numpy as np    
    
    from . import db, tile_mosaic
        
    np.set_printoptions(precision=6)
    
    #field='cos'
    
    # tile_npix = 4096
    # pscale = 0.1
    # 
    # ra, dec = 150.1, 2.29
    # rsize = 48
    # overlap = 0
    # 
    # # New tiles with overlaps
    # ra, dec = 150.125, 2.2
    # rsize = 45
    # pscale = 0.080
    # tile_npix = 2048+256
    
    if 0:
        ra, dec, rsize = 34.40869, -5.16299, 30
        field = 'uds'
        pscale = 0.08
        tile_npix = 2048+256
        
        
    tile_arcmin = tile_npix*pscale/60
    overlap=tile_arcmin/9
    
    print(f'{field}: tile size, arcmin = {tile_arcmin}')
    
    tiles = define_tiles(ra=ra, dec=dec, 
                         size=(rsize*2, rsize*2), 
                         tile_size=tile_arcmin, 
                         overlap=overlap, field=field, 
                         pixscale=pscale, theta=0)
        
    rows = []
    for t in tiles:
        wcs = tiles[t]
        sr = utils.SRegion(wcs.calc_footprint())
                    
        wcsh = utils.to_header(wcs)
        row = [t]
        for k in wcsh:
            row.append(wcsh[k])
        
        try:
            fp = sr.polystr()[0]
        except:
            fp = sr.xy[0].__str__().replace('\n', ',').replace('   ', ',')
            fp = fp.replace(' -',',-')
            fp = fp.replace(' ', '').replace('[','(')
            fp = fp.replace(']', ')').replace(',,',',')

        row.append(fp)     
        
        rows.append(row)
    
    names = ['tile']
    for k in wcsh:
        names.append(k.lower())
    
    names.append('footprint')

    tiles = utils.GTable(rows=rows, names=names)
    tiles['status'] = initial_status
    
    tiles['field'] = field
    
    # Offset of new WCS definition
    tiles['crpix1'] -= 0.5
    tiles['crpix2'] -= 0.5
    
    for c in ['cd1_2', 'cd2_1']:
        if c in tiles.colnames:
            tiles.remove_column(c)
            
    #db.send_to_database('cosmos_tiles', cos_tile, if_exists='replace')
    if send_to_database:
        db.execute(f"delete from combined_tiles where field = '{field}'")
        db.send_to_database('combined_tiles', tiles, if_exists='append')
    
    return tiles
    
def split_tiles(root='abell2744-080-08.08', ref_tile=(8,8), filters=['visr','f125w','h'], optical=False, suffix='.rgb', xsize=32, zoom_levels=[4,3,2,1], force=False, scl=1, invert=False, verbose=True, rgb_scl=[1,1,1], rgb_min=-0.01, make_combinations=True, norm_kwargs=None, pix_per_tile=2304, pl=2, pf=1):
    """
    Split image into 256 pixel tiles for map display
    """
    import matplotlib.pyplot as plt
    from skimage.io import imsave
    
    from grizli.pipeline import auto_script
    
    # nx = (2048+256)
    nx = pix_per_tile
    nsub = 2048

    if pix_per_tile >= 4096:
        nsub = 4096
        
    dpi = int(nx/xsize)
    
    if os.path.exists(f'{root}{suffix}.png') & (~force):
        return True
    
    try:
        _ = auto_script.field_rgb(root=root,
                                  xsize=xsize, filters=filters, 
                                  full_dimensions=2**optical, HOME_PATH=None, 
                                  add_labels=False,
                                  gzext='*', suffix=suffix, 
                                  output_format='png',
                                  scl=scl, invert=invert, 
                                  rgb_scl=rgb_scl,
                                  rgb_min=rgb_min,
                                  pl=pl,
                                  pf=pf,
                                  norm_kwargs=norm_kwargs)
    except IndexError:
        return False
    
    fig = _[-1]
    
    base = '_'.join(root.split('-')[:-1]).replace('+','_') + '_' + suffix[1:]
    
    tx, ty = np.asarray(root.split('-')[-1].split('.'),dtype=int)

    for iz, zoom in enumerate(zoom_levels):
        if iz > 0:
            zoom_img = f'{root}{suffix}.{2**iz:d}.png'
            # fig.savefig(zoom_img, dpi=dpi/2**iz)
            os.system(f'convert {root}{suffix}.png -scale {1/2**iz*100:.2f}% {root}{suffix}.{2**iz:d}.png')
            img = plt.imread(zoom_img)
        else:
            img = plt.imread(f'{root}{suffix}.png')
        
        if verbose:
            print(f'zoom: {zoom} {img.shape}')
        
        if img.ndim == 3:
            img = img[::-1,:,:]
        else:
            img = img[::-1,:]
        
        ntile = int(nsub/2**(4-zoom)/256)
        left = (tx - ref_tile[0])*ntile
        bot = -(ty - ref_tile[1])*ntile + (4096 // nsub) * ntile
        # print(zoom, ntile, left, bot)

        #axes[iz].set_xlim(-ntile*0.1, ntile*(1.1)-1)
        #axes[iz].set_ylim(*axes[iz].get_xlim())

        for i in range(ntile):
            xi = left + i
            for j in range(ntile):
                yi = bot - j - 1
                
                slx = slice((i*256), (i+1)*256)
                sly = slice((j*256), (j+1)*256)
                
                tile_file = f'{root}-tiles/{base}/{zoom}/{yi}/{xi}.png'
                if verbose > 1:
                    print(f'  {i} {j} {tile_file}')
                
                dirs = tile_file.split('/')
                for di in range(1,5):
                    dpath = '/'.join(dirs[:di])
                    #print(dpath)
                    if not os.path.exists(dpath):
                        os.mkdir(dpath)
                                     
                if img.ndim == 3:
                    imsave(tile_file, img[sly, slx, :][::-1,:,:],
                       plugin='pil', format_str='png')
                else:
                    imsave(tile_file, img[sly, slx][::-1,:],
                       plugin='pil', format_str='png')


def make_all_tile_images(root, force=False, ref_tile=(8,8), cleanup=True, zoom_levels=[4,3,2,1], brgb_filts=['visr','visb','uv'], rgb_filts=['visr','j','h'], blue_is_opt=True, make_opt_filters=True, make_ir_filters=True, make_combinations=True, rgb_only=False, pix_per_tile=2304, **kwargs):
    """
    Make tiles for map
    """
    import matplotlib.pyplot as plt
    
    from grizli.pipeline import auto_script
    
    #root = f'{field}-080-08.08'

    files = glob.glob(f'{root}-[hvuj]*')
    files += glob.glob(f'{root}*.rgb.png')
    
    all_files = glob.glob(f'{root}-f*_sci.fits*')
    all_files += glob.glob(f'{root}-clea*_sci.fits*')
    all_filters = [f.split(f'{root}-')[1].split('_dr')[0] for f in all_files]
    
    if len(files) == 0:
        auto_script.make_filter_combinations(root, 
                          filter_combinations={'h':['F140W','F160W'], 
                                               'j':['F105W','F110W','F125W'],
                                            'visr':['F850LP','F814W','F775W'],
                                     'visb':['F606W','F555W', 'F606WU'][:-1],
                                'uv':['F438WU','F435W','F435WU', 'F475W'], 
                                'sw':['F070W-CLEAR','F090W-CLEAR',
                                      'F115W-CLEAR','F150W-CLEAR',
                                      'F200W-CLEAR',
                                      'F182M-CLEAR','F210M-CLEAR',
                                      'F150W2-CLEAR',
                                  ],
                                'lw':['F277W-CLEAR','F356W-CLEAR',
                                      'F410M-CLEAR','F444W-CLEAR',
                                      'F322W2-CLEAR',
                                  ],
                            }, 
                                            weight_fnu=False)

    if 'j' in rgb_filts:
        rgb_scl = [1.1, 0.8, 8]
    else:
        rgb_scl = [1,1,8.]
        
    if make_combinations:
        split_tiles(root, ref_tile=ref_tile, 
                filters=rgb_filts, zoom_levels=zoom_levels,
                optical=False, suffix='.rgb', xsize=32, scl=1,
                force=force,
                rgb_scl=rgb_scl,
                pix_per_tile=pix_per_tile,
                pl=1., pf=1)

        plt.close('all')
    
    if (len(glob.glob(f'{root}*.brgb.png')) == 0) & make_combinations:
        split_tiles(root, ref_tile=ref_tile, 
                    filters=brgb_filts, zoom_levels=zoom_levels,
                    optical=blue_is_opt, suffix='.brgb', xsize=32, scl=8,
                    force=force, rgb_scl=[1., 1.2, 1.4],
                    rgb_min=-0.018,
                    pix_per_tile=pix_per_tile,
                    pl=2, pf=1)

        plt.close('all')
    
    # JWST SW
    if (len(glob.glob(f'{root}*.swrgb.png')) == 0) & make_combinations:
        filters = []
        for f in ['f090w-clear','f115w-clear','f150w-clear','f200w-clear',
                  'f182m-clear','f210m-clear']:
            if f in all_filters:
                filters.append(f)

        split_tiles(root, ref_tile=ref_tile, 
                    filters=filters,
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.swrgb', xsize=32, scl=2,
                    force=force, rgb_scl=[1,1.01,1.01], rgb_min=-0.018,
                    pix_per_tile=pix_per_tile,
                    pl=1, pf=1,
                    # norm_kwargs={'stretch': 'asinh', 'min_cut': -0.01,
                    #              'max_cut': 1.0, 'clip':True,
                    #              'asinh_a':0.03},
                    )

        plt.close('all')

    # JWST LW
    if (len(glob.glob(f'{root}*.lwrgb.png')) == 0) & make_combinations:
        filters = []
        for f in ['f277w-clear', 'f356w-clear', 'f444w-clear',
                  'f410m-clear', 'f365m-clear', 'f460m-clear',
                  'f480m-clear', 'f335m-clear', 'f300m-clear',
                  'f250m-clear', 'f360m-clear', 'f430m-clear',
                  ]:
            if f in all_filters:
                filters.append(f)
                
        split_tiles(root, ref_tile=ref_tile, 
                    filters=filters,
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.lwrgb', xsize=32, scl=4,
                    force=force, rgb_scl=[1.4,1.2,1.01],
                    # norm_kwargs={'stretch': 'asinh', 'min_cut': -0.01,
                    #              'max_cut': 1.0, 'clip':True,
                    #              'asinh_a':0.03},
                    pl=2, pf=1,
                    pix_per_tile=pix_per_tile,
                    rgb_min=-0.018)

        plt.close('all')
    
    # JWST NIRCam MB
    if (len(glob.glob(f'{root}*.mb1um.png')) == 0) & make_combinations:
        filters = []
        for f in ['f140m-clear','f182m-clear','f210m-clear']:
            if f in all_filters:
                filters.append(f)

        if len(filters) == 3:
            split_tiles(root, ref_tile=ref_tile, 
                    filters=filters,
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.mb1um', xsize=32, scl=2,
                    force=force, rgb_scl=[1,1.01,1.01], rgb_min=-0.018,
                    pix_per_tile=pix_per_tile,
                    pl=1, pf=1,
                    # norm_kwargs={'stretch': 'asinh', 'min_cut': -0.01,
                    #              'max_cut': 1.0, 'clip':True,
                    #              'asinh_a':0.03},
                    )

            plt.close('all')
    
    if (len(glob.glob(f'{root}*.mb3um.png')) == 0) & make_combinations:
        filters = []
        for f in ['f300m-clear','f335m-clear','f360m-clear']:
            if f in all_filters:
                filters.append(f)

        if len(filters) == 3:
            rgb_scl = [1.15,1.1,0.92]
            split_tiles(root, ref_tile=ref_tile, 
                    filters=filters,
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.mb3um', xsize=32, scl=4,
                    force=force, rgb_scl=rgb_scl, rgb_min=-0.018,
                    pix_per_tile=pix_per_tile,
                    pl=1, pf=1,
                    # norm_kwargs={'stretch': 'asinh', 'min_cut': -0.01,
                    #              'max_cut': 1.0, 'clip':True,
                    #              'asinh_a':0.03},
                    )

            plt.close('all')

    if (len(glob.glob(f'{root}*.mb4um.png')) == 0) & make_combinations:
        filters = []
        for f in ['f410m-clear','f430m-clear','f460m-clear']:
            if f in all_filters:
                filters.append(f)

        if len(filters) == 3:
            rgb_scl = [1.01, 1.01, 1.0]
            split_tiles(root, ref_tile=ref_tile, 
                    filters=filters,
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.mb4um', xsize=32, scl=4,
                    force=force, rgb_scl=rgb_scl, rgb_min=-0.018,
                    pl=1, pf=1,
                    pix_per_tile=pix_per_tile,
                    # norm_kwargs={'stretch': 'asinh', 'min_cut': -0.01,
                    #              'max_cut': 1.0, 'clip':True,
                    #              'asinh_a':0.03},
                    )

            plt.close('all')
    
    # All NIRCam - prefer r, g, b = F444W, F277W, F115W
    # but look for other combinations if that not available
    if (len(glob.glob(f'{root}*.ncrgb.png')) == 0) & make_combinations:
        filters = []
        for f in ['f444w-clear','f410m-clear','f322w2-clear']:
            if f in all_filters:
                filters.append(f)
                break
        for f in ['f277w-clear','f356w-clear','f300m-clear']:
            if f in all_filters:
                filters.append(f)
                break
        for f in ['f115w-clear','f090w-clear','f150w-clear','f200w-clear','f150w2-clear']:
            if f in all_filters:
                filters.append(f)
                break
        
        rgb_scl = [1.5, 0.8, 1]
        
        if ('cos' in root) & ('f150w2-clear' in all_filters) & (len(filters) < 3):
            filters = ['f814w','f150w2-clear','f444w-clear']
        elif ('sextans' in root):
            filters = ['f115w-clear','f200w-clear','f360m-clear']
            rgb_scl = [1.3, 0.8, 1.02]
        elif ('ulasj1342' in root):
            filters = ['f115w-clear','f250m-clear','f430m-clear']
            rgb_scl = [1.3, 0.8, 1.02]
        elif ('j031124m5823' in root):
            filters = ['f200w-clear','f300m-clear','f444w-clear']
            rgb_scl = [1.2, 1.05, 1.0]
        elif 'spiderweb' in root:
            filters = ['f115w-clear','f182m-clear','f410m-clear']
            rgb_scl = [1.3, 0.8, 1.02]
        elif 'aspire' in root:
            filters = ['f115w-clear','f200w-clear','f356w-clear']
            rgb_scl = [1.3, 0.8, 1.02]
        elif 'eiger' in root:
            filters = ['f115w-clear','f200w-clear','f356w-clear']
            rgb_scl = [1.3, 0.8, 1.02]
        elif 'nexus-j1753' in root:
            filters = ['f115w-clear','f200w-clear','f444w-clear']
            rgb_scl = [1.0, 0.9, 1.01]

        if root.split('-')[0] in ['cluster','eiger','aspire']:
            norm_kwargs=None
            scl = 2
            rgb_min = -0.01
        else:
            scl = 4
            rgb_min = -0.018
            norm_kwargs={'stretch': 'asinh', 'min_cut': -0.01,
                         'max_cut': 1.0, 'clip':True,
                         'asinh_a':0.03}

        split_tiles(root, ref_tile=ref_tile, 
                    filters=filters,
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.ncrgb', xsize=32, scl=scl,
                    force=force, rgb_min=rgb_min,
                    rgb_scl=rgb_scl,
                    norm_kwargs=norm_kwargs,
                    pix_per_tile=pix_per_tile,
                    pl=1.5, pf=1)

        plt.close('all')
    
    if ('dracoii' in root) & make_combinations:
        
        split_tiles(root, ref_tile=ref_tile, 
                    filters=[f.lower() for f in ['F090W-CLEAR',
                             'F150W-CLEAR']],
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.swrgb', xsize=32, scl=4,
                    pix_per_tile=pix_per_tile,
                    force=force, rgb_scl=[1,1,1], rgb_min=-0.018)

        plt.close('all')

        split_tiles(root, ref_tile=ref_tile, 
                    filters=[f.lower() for f in ['F360M-CLEAR',
                             'F480M-CLEAR']],
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.lwrgb', xsize=32, scl=4,
                    pix_per_tile=pix_per_tile,
                    force=force, rgb_scl=[1,1,1], rgb_min=-0.018)

        plt.close('all')

        split_tiles(root, ref_tile=ref_tile, 
                    filters=[f.lower() for f in ['F090W-CLEAR','F150W-CLEAR',
                                                 'F480M-CLEAR',
                                                ]],
                    zoom_levels=zoom_levels,
                    optical=True, suffix='.ncrgb', xsize=32, scl=4,
                    pix_per_tile=pix_per_tile,
                    force=force, rgb_scl=[1.4, 0.6, 0.35], rgb_min=-0.018)

        plt.close('all')
        
    if root.startswith('cos') & make_combinations:
        if len(glob.glob(f'{root}*.vi.png')) == 0:
            split_tiles(root, ref_tile=ref_tile, 
                        filters=['f814w','f160w'], zoom_levels=zoom_levels,
                        optical=False, suffix='.vi', xsize=32, scl=0.8,
                        pix_per_tile=pix_per_tile,
                        force=force, rgb_scl=[1, 1, 1], rgb_min=-0.018)

            plt.close('all')
    
    if rgb_only:
        make_ir_filters = make_opt_filters = False
        
    ####### Single filters
    # IR
    if make_ir_filters:
        files = []
        for ir_filt in ['f098m','f105w','f110w','f125w','f140w','f160w']:
            files += glob.glob(f'{root}-{ir_filt}*sci.fits*')
        
        # MIRI on 80mas 
        for ir_filt in ['f560w','f770w','f1000w','f1130w',
                        'f1280w','f1500w','f1800w','f2100w','f2550w']:
            files += glob.glob(f'{root}-{ir_filt}*sci.fits*')
        
        files.sort()
        filts = [('-'.join(file.split(f'{root}-')[1:])).split('_')[0]
                 for file in files]
                 
        for filt in filts:
            if os.path.exists(f'{root}.{filt}.png'):
                continue

            split_tiles(root, ref_tile=ref_tile, 
                    filters=[filt], zoom_levels=zoom_levels,
                    optical=False, suffix=f'.{filt}', xsize=32, 
                    pix_per_tile=pix_per_tile,
                    force=force, scl=2, invert=True)

            plt.close('all')
    
    if make_opt_filters:
        # Optical, 2X pix
        files = glob.glob(f'{root}-f[2-8]*sci.fits*')
        files += glob.glob(f'{root}-f*clear*sci.fits*')
        files.sort()
        filts = [file.split(f'{root}-')[1].split('_')[0] for file in files]
        for filt in filts:
            if os.path.exists(f'{root}.{filt}.png'):
                continue

            split_tiles(root, ref_tile=ref_tile, 
                filters=[filt], zoom_levels=zoom_levels,
                optical=True, suffix=f'.{filt}', xsize=32, 
                pix_per_tile=pix_per_tile,
                force=force, scl=4, invert=True)

            plt.close('all')
    
    if cleanup:
        files = glob.glob(f'{root}-[vhuj]*fits*')
        files.sort()
        for file in files:
            print(f'rm {file}')
            os.remove(file)


TILE_FILTERS = ['F336W', 'F350LP', 'F390W', 'F435W', 'F438W', 'F475W',
                'F555W', 'F606W', 'F625W', 'F775W', 
                'F814W', 'F850LP', 
                'F098M', 'F105W', 'F110W', 'F125W', 'F140W', 'F160W',
                'F070W-CLEAR','F090W-CLEAR','F115W-CLEAR','F150W-CLEAR', # NRC
                'F200W-CLEAR','F210M-CLEAR','F250M-CLEAR',
                'F150W2-CLEAR','F322W2-CLEAR',
                'F277W-CLEAR','F356W-CLEAR','F444W-CLEAR',
                'F300M-CLEAR','F360M-CLEAR','F410M-CLEAR', 'F480M-CLEAR',
                'CLEAR-F115W','CLEAR-F150W','CLEAR-F200W', # NIS
                'F560W','F770W','F1000W','F1280W','F1500W', # MIR
                'F1800W','F2100W','F2550W',
                ][1:]


def process_tile_filter(field='cos', tile='01.01', filter='F444W-CLEAR', fetch_existing=False, make_tile_images=True, make_combinations=False, **kwargs):
    """
    Run `process_tile` for a single filter
    """
    from grizli.aws import visit_processor, db

    NOW = astropy.time.Time.now().mjd
    
    print(f'Set status=1 in `combined_tiles_filters` for {field} {tile} {filter}')
    
    db.execute(f"""update combined_tiles_filters
    set status=1, modtime={NOW}
    where tile = '{tile}' and field='{field}' and filter='{filter}'
    """)
    
    try:
        status = process_tile(field=field, tile=tile, filters=[filter], 
                         fetch_existing=fetch_existing,
                         make_tile_images=make_tile_images,
                         make_combinations=make_combinations,
                         **kwargs)
    except:
        status = 10
    
    NOW = astropy.time.Time.now().mjd

    if (status == 2):
        # Set status = 3 in parent table to indicate that it should be rerun
        # to generate the RGB images
        print(f'Set status=3 in `combined_tiles` for {field} {tile}')
        
        db.execute(f"""update combined_tiles
        set status=3
        where tile = '{tile}' and field='{field}'
        """)
    
    # update status in filter table
    print(f'Set status={status} in `combined_tiles_filters` for {field} {tile} {filter}')
    
    db.execute(f"""update combined_tiles_filters
    set status={status}, modtime={NOW}
    where tile = '{tile}' and field='{field}' and filter='{filter}'
    """)
    return status


def process_tile(field='cos', tile='01.01', filters=TILE_FILTERS, fetch_existing=True, cleanup=True, make_catalog=False, clean_flt=True, pixfrac=0.75, make_tile_images=True, make_combinations=True, rgb_only=False, **kwargs):
    """
    
    Returns
    status : int
        - 2 - completed successfully
        - 4 - no images created
        - 10 - catalog requested but failed to create
    """
    import numpy as np
    
    import astropy.io.fits as pyfits
    import astropy.wcs as pywcs
    from grizli.aws import visit_processor, db
    from grizli.pipeline import auto_script
    from grizli import prep
    
    if field in ('sextansa','cos','abell370','abell2744',
                 'macs0416','macs1149','macs1423','macs0417',
                 'egs','uds','gds','gdn','egs-v2',
                 'ulasj1342'):
        make_catalog = False
    
    if field.startswith('aspire'):
        make_catalog = False
    
    utils.LOGFILE = 'mosaic.log'
    
    print(f'Set status=1 in `combined_tiles` for {field} {tile}')
    db.execute(f"update combined_tiles set status=1 where tile = '{tile}' and field='{field}'")
    
    row = db.SQL(f"select * from combined_tiles where tile = '{tile}' and field='{field}'")
    
    h = pyfits.Header()
    for k in row.colnames:
        if k in ['footprint']:
            continue
        
        h[k.upper()] = row[k][0]
        
    ir_wcs = pywcs.WCS(h)
    
    root = f'{field}-080-{tile}'
    
    if fetch_existing:
        os.system(f"""aws s3 sync s3://grizli-v2/ClusterTiles/{field}/ ./
                          --exclude "*"
                          --include "{root}*_dr?*fits.gz"
                          --exclude "{root}*-f770w*"
                          --exclude "{root}*-f1???w*"
                          --exclude "{root}*-f2???w*"
                          """.replace('\n', ' '))
        
        files = glob.glob(f'{root}*gz')
        for file in files:
            os.system(f'gunzip --force {file}')
            
    if not rgb_only:
        visit_processor.cutout_mosaic(
            rootname=root,
            skip_existing=True,
            ir_wcs=ir_wcs,
            filters=filters,
            pixfrac=pixfrac,
            kernel='square',
            s3output=None,
            gzip_output=False,
            clean_flt=clean_flt
        )
    
    files = glob.glob(f'{root}*_dr*fits*')
    if len(files) == 0:
        db.execute(f"update combined_tiles set status=4 where tile = '{tile}'")
        print(f'No images found, set status=4 in `combined_tiles` for {field} {tile}')
        
        return 4
    
    if make_catalog:
        import golfir.catalog
        
        golfir.catalog.make_charge_detection(
            root,
            ext='ir',
            filters=[
                'f160w', 'f140w', 'f125w', 'f110w', 'f105w',
                'f814w', 'f850lp',
                'f277w-clear','f356w-clear','f444w-clear'
            ],
            use_hst_kernel=False,
            subtract_background=True
        )

        phot = auto_script.multiband_catalog(
            field_root=root,
            phot_apertures=prep.SEXTRACTOR_PHOT_APERTURES_ARCSEC[:4]
        ) #, **phot_kwargs)

        if len(phot) == 0:
            # Empty catalog
            db.execute(
                "update combined_tiles set status=10 "
                + f"where tile = '{tile}' AND field = '{field}'"
            )
        
            if cleanup:

                files = glob.glob(f'{root}*')
                files.sort()
                for file in files:
                    print(f'rm {file}')
                    os.remove(file)
        
            return 10
        
        for i in [4,5,6]:
            for c in phot.colnames:
                if c.endswith(f'{i}'):
                    phot.remove_column(c)
    
        for c in phot.colnames:
            if c in ['ra','dec','x_world','y_world']:
                continue
            
            if phot[c].dtype == np.float64:
                phot[c] = phot[c].astype(np.float32)
            elif phot[c].dtype == np.int64:
                phot[c] = phot[c].astype(np.int32)
              
        phot['tile'] = tile
        phot['field'] = field
    
        if 'id' in phot.colnames:
            phot.remove_column('id')
    
        for c in ['xmin','xmax','ymin','ymax']:
            if c in phot.colnames:
                phot.rename_column(c, c+'pix')
    
        for c in phot.colnames:
            if '-' in c:
                phot.rename_column(c, c.replace('-','_'))
            
        if 0:
            db.execute('CREATE TABLE combined_tile_phot AS SELECT * FROM cosmos_tile_phot limit 0')
            db.execute('ALTER TABLE combined_tile_phot DROP COLUMN id;')
            db.execute('ALTER TABLE combined_tile_phot ADD COLUMN id SERIAL PRIMARY KEY;')
        
        db.execute(f"DELETE from combined_tile_phot WHERE tile='{tile}' and field='{field}'")
    
        seg = pyfits.open(f'{root}-ir_seg.fits')
    
        ### IDs on edge
        edge = np.unique(seg[0].data[16:19,:])
        edge = np.append(edge, np.unique(seg[0].data[-19:-16,:]))
        edge = np.append(edge, np.unique(seg[0].data[:, 16:19]))
        edge = np.append(edge, np.unique(seg[0].data[:, -19:-16]))
        edge = np.unique(edge)
        phot['edge'] = np.in1d(phot['number'], edge)*1
    
        ### Add missing columns
        cols = db.SQL('select * from combined_tile_phot limit 2').colnames
        for c in phot.colnames:
            if c not in cols:
                print('Add column {0} to `combined_tile_phot` table'.format(c))
                if phot[c].dtype in [np.float64, np.float32]:
                    SQL = "ALTER TABLE combined_tile_phot ADD COLUMN {0} real;".format(c)
                else:
                    SQL = "ALTER TABLE combined_tile_phot ADD COLUMN {0} int;".format(c)
                
                db.execute(SQL)
            
        db.send_to_database('combined_tile_phot', phot, if_exists='append')
    
        if 'id' not in cols:
            # Add unique id index column
            db.execute('ALTER TABLE combined_tile_phot ADD COLUMN id SERIAL PRIMARY KEY;')
    
        # Use db id
        ids = db.SQL(f"SELECT number, id, ra, dec, tile from combined_tile_phot WHERE tile='{tile}' AND field='{field}'")
        idx, dr = ids.match_to_catalog_sky(phot)
        
        phot['id'] = phot['number']
        for c in ['xmin','xmax','ymin','ymax']:
            if c+'pix' in phot.colnames:
                phot.rename_column(c+'pix', c)
    
        golfir.catalog.switch_segments(seg[0].data, phot, ids['id'][idx])
        pyfits.writeto(f'{root}-ir_seg.fits', data=seg[0].data, 
                       header=seg[0].header, overwrite=True)
        
        # zip and copy
        drz_files = glob.glob(f'{root}-ir_dr*fits')
        drz_files += glob.glob(f'{root}*seg.fits')
        drz_files.sort()
        
        for file in drz_files:
            cmd = f'gzip --force {file}'
            print(cmd)
            os.system(cmd)
        
        os.system(f'aws s3 sync ./ s3://grizli-v2/ClusterTiles/{field}/' + 
                  f' --exclude "*" --include "{root}*gz" --include "{root}_phot.fits"'
                  ' --acl public-read')
        
    if make_tile_images:
        ### Make subtiles
        ref_tiles = {'cos': (16,16), 
                     'uds': (11, 10),
                     'abell2744': (8, 8), 
                     'egs': (10,14),
                     'gds': (9,9), 
                     'gdn': (9,9),
                     'sgas1723': (4,4),
                     'sgas1226': (2,2),
                     'smacs0723': (4,4), 
                     'dracoii':(5,4),
                     'egs-v2': (10,14),
                     'j1235':(6,4),
                     'whl0137':(5,5),
                     'macs1423':(8,8),
                     'macs0416':(8,8),
                     'macs0417':(8,8),
                     'macs1423':(8,8),
                     'abell370':(8,8),
                     'macs1149':(8,8),
                     }
    
        ref_tileq = db.SQL(f"""select * from combined_tiles where field = '{field}'
                AND crpix1 > 0 AND crpix2 > 0
                order by (POW(crpix1,2) + POW(crpix2,2)) limit 1""")
        
        pix_per_tile = ref_tileq['naxis1'][0]
        
        if field in ref_tiles:
            ref_tile = ref_tiles[field]
        else:
            # ref_tile = (9, 9)
            if len(ref_tileq) == 1:
                ref_tile = tuple(np.array(ref_tileq['tile'][0].split('.')).astype(int))
            else:
                ref_tile = (9, 9)
        
        print(f'field {field}, ref_tile: {ref_tile}')
        
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            
            make_all_tile_images(root, force=False, ref_tile=ref_tile,
                                 rgb_filts=['h','j','visr'],
                                 brgb_filts=['visr','visb','uv'],
                                 blue_is_opt=(field not in ['j013804m2156']), 
                                 make_combinations=make_combinations,
                                 rgb_only=rgb_only,
                                 pix_per_tile=pix_per_tile,
                                 **kwargs)
    
    dirs = glob.glob(f'./{root}-tiles/*')
    dirs.sort()
    
    for d in dirs:
        target = os.path.basename(d)
        print(f'Sync {d} >> s3://grizli-v2/ClusterTiles/Map/{field}/{target}/')
    
        os.system(f'aws s3 sync {d}/ ' + 
                  f' s3://grizli-v2/ClusterTiles/Map/{field}/{target} ' + 
                  '--acl public-read --quiet')
                      
    if not rgb_only:
        ### Gzip products
        drz_files = glob.glob(f'{root}-*_dr*fits')
        drz_files += glob.glob(f'{root}*seg.fits')
        drz_files.sort()
        
        for file in drz_files:
            cmd = f'gzip --force {file}'
            print(cmd)
            os.system(cmd)
        
        os.system(f'aws s3 sync ./ s3://grizli-v2/ClusterTiles/{field}/' + 
                  f' --exclude "*" --include "{root}*gz"' +
                  f' --include "{root}*wcs.csv"' +
                  f' --include "{root}*fp.png" --acl public-read')
    
    db.execute(f"update combined_tiles set status=2 where tile = '{tile}' AND field = '{field}'")
    print(f'Set status=2 in `combined_tiles` for {field} {tile}')
    
    if cleanup:
        print(f'rm -rf {root}-tiles')
        os.system(f'rm -rf {root}-tiles')
        
        files = glob.glob(f'{root}*')
        files.sort()
        for file in files:
            print(f'rm {file}')
            os.remove(file)
    
    return 2



def get_random_tile():
    """
    Find a visit that needs processing
    """
    from grizli.aws import db
    
    all_tiles = db.SQL(f"""SELECT tile, field 
                           FROM combined_tiles
                           WHERE status=0""")
    
    if len(all_tiles) == 0:
        return None, None
    
    tile, field = all_tiles[np.random.randint(0, len(all_tiles))]
    return tile, field


def run_one(own_directory=True, rgb_only=True, make_catalog=False, **kwargs):
    """
    Run a single random visit
    """
    import os
    import time
    from grizli.aws import db

    ntile = db.SQL("""SELECT count(status)
                       FROM combined_tiles
                       WHERE status = 0""")['count'][0] 
    
    tile, field = get_random_tile()
    if tile is None:
        with open(os.path.join(HOME, 'tile_finished.txt'),'w') as fp:
            fp.write(time.ctime() + '\n')
    else:
        print(f'============  Run tile  ==============')
        print(f'{field} {tile}')
        print(f'========= {time.ctime()} ==========')
        
        with open(os.path.join(HOME, 'tile_history.txt'),'a') as fp:
            fp.write(f'{time.ctime()} {tile}\n')
        
        if own_directory:
            path = os.path.join(HOME, f'{field}-{tile}')
            if not os.path.exists(path):
                os.makedirs(path)
                
            os.chdir(path)
            
        #process_visit(tile, clean=clean, sync=sync)
        process_tile(tile=tile, field=field, rgb_only=rgb_only, 
                     make_catalog=make_catalog,
                     **kwargs)


def get_random_tile_filter():
    """
    Find a visit that needs processing
    """
    from grizli.aws import db
    
    all_tiles = db.SQL(f"""SELECT tile, field, filter
                           FROM combined_tiles_filters
                           WHERE status=0""")
    
    if len(all_tiles) == 0:
        return None, None, None
    
    tile, field, filt = all_tiles[np.random.randint(0, len(all_tiles))]
    return tile, field, filt


def run_one_filter_tile(own_directory=True, **kwargs):
    """
    Run a single random visit
    """
    import os
    import time
    from grizli.aws import db

    ntile = db.SQL("""SELECT count(status)
                       FROM combined_tiles_filters
                       WHERE status = 0""")['count'][0] 
    
    tile, field, filt = get_random_tile_filter()
    if tile is None:
        with open(os.path.join(HOME, 'tile_filter_finished.txt'),'w') as fp:
            fp.write(time.ctime() + '\n')
    else:
        print(f'============  Run tile filter ========')
        print(f'{field} {tile} {filt}')
        print(f'========= {time.ctime()} ==========')
        
        with open(os.path.join(HOME, 'tile_filter_history.txt'),'a') as fp:
            fp.write(f'{time.ctime()} {tile}\n')
        
        if own_directory:
            path = os.path.join(HOME, f'{field}-{tile}-{filt}')
            if not os.path.exists(path):
                os.makedirs(path)
                
            os.chdir(path)
            
        #process_visit(tile, clean=clean, sync=sync)
        process_tile_filter(tile=tile, field=field, filter=filt, **kwargs)


def launch_ec2_instances(nmax=50, count=None, templ='lt-0e8c2b8611c9029eb,Version=25'):
    """
    Launch EC2 instances from a launch template that run through all 
    status=0 associations/tiles and then terminate
    
    Version 19 is the latest run_all_visits.sh
    Version 20 is the latest run_all_tiles.sh
    Version 24 is run_all_visits with a new python39 environment
    
    Version 25 is run_all_tile_filters.sh with a new python39 environment
    
    """
    from grizli.aws import db
    
    if count is None:
        assoc = db.SQL("""SELECT tile, field, filter
                      FROM combined_tiles_filters
                      WHERE status = 0
                      GROUP BY tile, field, filter""")
    
        count = int(np.minimum(nmax, len(assoc)/2))

    if count == 0:
        print('No associations to run, abort.')
        return True
    else:
        print(f'# Launch {count} instances with LaunchTemplateId={templ}: ')
        cmd = f'aws ec2 run-instances --count {count}'
        cmd += f' --launch-template LaunchTemplateId={templ}'
        print(cmd)
        os.system(cmd)

def create_mosaic_from_tiles(assoc, filt='ir', clean=True):
    """
    Get tiles overlapping visit footprint
    """
    import glob    
    import astropy.io.fits as pyfits
    import astropy.table
    from grizli.aws import db
    
    olap_tiles = db.SQL(f"""SELECT tile, field
                        FROM combined_tiles t, exposure_files e
                        WHERE e.assoc = '{assoc}'
                        AND polygon(e.footprint) && polygon(t.footprint)
                        GROUP BY tile, field""")
    
    tx = np.array([int(t.split('.')[0]) for t in olap_tiles['tile']])
    ty = np.array([int(t.split('.')[1]) for t in olap_tiles['tile']])
    
    xm = tx.min()
    ym = ty.min()
    
    nx = tx.max()-tx.min()+1
    ny = ty.max()-ty.min()+1
    
    print(olap_tiles)
    
    field = olap_tiles['field'][0]
    for t in olap_tiles['tile']:
        # print(f'Fetch tile s3://grizli-v2/ClusterTiles/{field}/{field}-080-{t}-{filt}*')
        # 
        # os.system(f"aws s3 cp s3://grizli-v2/ClusterTiles/{field}/{field}-080-{t}-{filt}_drz_sci.fits.gz . ")
        # os.system(f"aws s3 cp s3://grizli-v2/ClusterTiles/{field}/{field}-080-{t}-{filt}_drz_wht.fits.gz . ")
        # os.system(f"aws s3 cp s3://grizli-v2/ClusterTiles/{field}/{field}-080-{t}-{filt}_seg.fits.gz . ")
        
        os.system(f"""aws s3 sync s3://grizli-v2/ClusterTiles/{field}/ ./
                          --exclude "*"
                          --include "{field}*{t}-{filt}*_dr?*fits.gz"
                          --include "{field}*{t}-ir_seg.fits.gz"
                          """.replace('\n', ' '))
        
    wcs = db.SQL(f"""SELECT * FROM combined_tiles 
                     WHERE tile = '{xm:02d}.{ym:02d}'
                     AND field = '{field}'""")
    
    h = pyfits.Header()
    for c in wcs.colnames:
        if c in ['footprint']:
            continue
        
        h[c] = wcs[c][0]
    
    xnpix = h['NAXIS1']
    ynpix = h['NAXIS2']
    
    olap = 256
    pad = 32
    
    h['NAXIS1'] = h['NAXIS1']*nx - olap*(nx-1)
    h['NAXIS2'] = h['NAXIS2']*ny - olap*(ny-1)
    
    img_shape = (h['NAXIS2'], h['NAXIS1'])
    sci = np.zeros(img_shape, dtype=np.float32)
    wht = np.zeros(img_shape, dtype=np.float32)
    seg = np.zeros(img_shape, dtype=int)
    
    skip_tiles = []
    
    for tile, txi, tyi in zip(olap_tiles['tile'], tx, ty):
        _file = f'{field}-080-{txi:02d}.{tyi:02d}-{filt}_dr*_sci.fits.gz'
        _files = glob.glob(_file)
        if len(_files) == 0:
            msg = f'*ERROR* {_file} not found'
            utils.log_comment(utils.LOGFILE, msg, verbose=True)
            continue
            
        file = _files[0]
        
        msg = f'Add tile {file} to {assoc} mosaic'
        utils.log_comment(utils.LOGFILE, msg, verbose=True)

        im = pyfits.open(file)
        if im[0].shape[0] != xnpix:
            print(f'Bad size ({im[0].data.shape})')
            skip_tiles.append(tile)
            continue
            
        olap_dx = (txi-xm)*olap
        olap_dy = (tyi-ym)*olap
        
        slx = slice((txi-xm)*xnpix - olap_dx + pad,
                    (txi-xm+1)*xnpix - olap_dx)
        sly = slice((tyi-ym)*ynpix - olap_dy + pad,
                    (tyi-ym+1)*ynpix - olap_dy)
        
        for k in im[0].header:
            if k not in h:
                h[k] = im[0].header[k]
                
        sci[sly, slx] = im[0].data[pad:, pad:]*1

        im = pyfits.open(file.replace('_sci','_wht'))
        wht[sly, slx] = im[0].data[pad:, pad:]**1

        im = pyfits.open(file.replace('_drz_sci','_seg'))
        seg[sly, slx] = im[0].data[pad:, pad:]**1
    
    _hdu = pyfits.PrimaryHDU(data=sci, header=h)
    _hdu.writeto(f'{assoc}-{filt}_drz_sci.fits', overwrite=True)
    
    _hdu = pyfits.PrimaryHDU(data=wht, header=h)
    _hdu.writeto(f'{assoc}-{filt}_drz_wht.fits', overwrite=True)
    
    _hdu = pyfits.PrimaryHDU(data=seg, header=h)
    _hdu.writeto(f'{assoc}-ir_seg.fits', overwrite=True)
    
    if clean:
        for tile, txi, tyi in zip(olap_tiles['tile'], tx, ty):
            files = glob.glob(f'cos-080-{txi:02d}.{tyi:02d}-{filt}*')
            for file in files:
                msg = f'rm {file}'
                utils.log_comment(utils.LOGFILE, msg, verbose=True)
                os.remove(file)

    #### Catalog
    cols = ['id','thresh', 'npix', 'tnpix',
            'xminpix as xmin', 'xmaxpix as xmax',
            'yminpix as ymin', 'ymaxpix as ymax', 
            'x', 'y', 'x2_image', 'y2_image', 'xy_image',
            'errx2', 'erry2', 'errxy',
            'a_image', 'b_image', 'theta_image',
            'cxx_image', 'cyy_image', 'cxy_image',
            'cflux', 'flux',
            'cpeak', 'peak', 'xcpeak', 'ycpeak', 'xpeak', 'ypeak',
            'flag',
            'x_image', 'y_image',
            #'x_world', 'y_world',
            'ra as x_world', 'dec as y_world', 
            'flux_iso', 'fluxerr_iso', 'area_iso', 'mag_iso', 'kron_radius',
            'kron_rcirc', 'flux_auto', 'fluxerr_auto', 'bkg_auto',
            'flag_auto', 'area_auto', 'flux_radius_flag', 'flux_radius_20',
            'flux_radius', 'flux_radius_90', 'tot_corr', 'mag_auto',
            'magerr_auto', 'flux_aper_0', 'fluxerr_aper_0', 'flag_aper_0',
            'bkg_aper_0', 'mask_aper_0', 'flux_aper_1', 'fluxerr_aper_1',
            'flag_aper_1', 'bkg_aper_1', 'mask_aper_1', 'flux_aper_2',
            'fluxerr_aper_2', 'flag_aper_2', 'bkg_aper_2', 'mask_aper_2',
            'flux_aper_3', 'fluxerr_aper_3', 'flag_aper_3', 'bkg_aper_3',
            'mask_aper_3']
    
    scols = ','.join(cols)
    
    tabs = []
    for tile, txi, tyi in zip(olap_tiles['tile'], tx, ty):
        if tile in skip_tiles:
            continue
            
        _SQL = f"""SELECT {scols} from combined_tile_phot
                   where tile = '{tile}'
                   AND x < 2048 AND y < 2048
                   AND x > {pad} AND y > {pad}"""
                   
        tab = db.SQL(_SQL)
                
        # Pixel offsets
        olap_dx = (txi-xm)*olap
        olap_dy = (tyi-ym)*olap

        dx = (txi-xm)*(2048+olap) - olap_dx
        dy = (tyi-ym)*(2048+olap) - olap_dx
        
        for c in tab.colnames:
            if c in ['xmin', 'xmax', 'x', 'x_image', 'xpeak', 'xcpeak']:
                tab[c] += dx
            elif c in ['ymin', 'ymax', 'y', 'y_image', 'ypeak', 'ycpeak']:
                tab[c] += dy
        
        tabs.append(tab)
        
        msg = f'{assoc}: Query {tile} catalog N={len(tab)} dx={dx} dy={dy}'
        utils.log_comment(utils.LOGFILE, msg, verbose=True)
    
    tab = astropy.table.vstack(tabs)
    for c in tab.colnames:
        tab.rename_column(c, c.upper())
    
    # Make it look as expected for grizli model
    if hasattr(tab['MAG_AUTO'], 'filled'):
        tab['MAG_AUTO'] = tab['MAG_AUTO'].filled(99)
        
    tab.rename_column('ID','NUMBER')
    tab['NUMBER'] = tab['NUMBER'].astype(int)
    tab.write(f'{assoc}-ir.cat.fits', overwrite=True)
    
    ### Grism
    if 0:
        os.system(f'aws s3 sync s3://grizli-v2/HST/Pipeline/{assoc}/Prep/ ./ --exclude "*" --include "*flt.fits" --include "*yml"')
    
        # grism_files = glob.glob('iehn*[a-p]_flt.fits')
        # grism_files.sort()
        #     
        # grp = auto_script.grism_prep(field_root=assoc, 
        #                              gris_ref_filters={'G141':['ir']},
        #                              files=grism_files,
        #                              refine_mag_limits=[18,23], 
        #                              PREP_PATH='./')
        # 
        # if len(glob.glob(f'{assoc}*_grism*fits*')) == 0:
        #     grism_files = glob.glob('*GrismFLT.fits')
        #     grism_files.sort()
        # 
        #     catalog = glob.glob(f'{assoc}-*.cat.fits')[0]
        #     try:
        #         seg_file = glob.glob(f'{assoc}-*_seg.fits')[0]
        #     except:
        #         seg_file = None
        # 
        #     grp = multifit.GroupFLT(grism_files=grism_files, direct_files=[], 
        #                             ref_file=None, seg_file=seg_file, 
        #                             catalog=catalog, cpu_count=-1, sci_extn=1, 
        #                             pad=256)
        # 
        #     # Make drizzle model images
        #     grp.drizzle_grism_models(root=assoc, kernel='point', scale=0.15)
        # 
        #     # Free grp object
        #     del(grp)
        # 
        # pline = auto_script.DITHERED_PLINE.copy()
        # args_file = f'{assoc}_fit_args.npy'
        # 
        # if (not os.path.exists(args_file)):
        # 
        #     msg = '# generate_fit_params: ' + args_file
        #     utils.log_comment(utils.LOGFILE, msg, verbose=True, show_date=True)
        # 
        #     pline['pixscale'] = 0.1 #mosaic_args['wcs_params']['pixel_scale']
        #     pline['pixfrac'] = 0.5  #mosaic_args['mosaic_pixfrac']
        #     if pline['pixfrac'] > 0:
        #         pline['kernel'] = 'square'
        #     else:
        #         pline['kernel'] = 'point'
        # 
        #     min_sens = 1.e-4
        #     min_mask = 1.e-4
        # 
        #     fit_trace_shift = True
        # 
        #     args = auto_script.generate_fit_params(field_root=assoc, prior=None, MW_EBV=0.0, pline=pline, fit_only_beams=True, run_fit=True, poly_order=7, fsps=True, min_sens=min_sens, min_mask=min_mask, sys_err=0.03, fcontam=0.2, zr=[0.05, 3.4], save_file=args_file, fit_trace_shift=fit_trace_shift, include_photometry=False, use_phot_obj=False)
        # 
        #     os.system(f'cp {args_file} fit_args.npy')

    



"""
Catalog table tools
"""
import os
import inspect

from collections import OrderedDict
import glob
import traceback

import numpy as np

import astropy.io.fits as pyfits
import astropy.units as u
import astropy.coordinates as coord
from astropy.table import Table

from . import utils

__all__ = [
    "table_to_radec",
    "table_to_regions",
    "randomize_segmentation_labels",
    "get_ukidss_catalog",
    "get_sdss_catalog",
    "get_twomass_catalog",
    "get_irsa_catalog",
    "get_gaia_radec_at_time",
    "get_gaia_DR2_vizier_columns",
    "get_gaia_vizier",
    "get_gaia_DR2_vizier",
    "gaia_dr2_conesearch_query",
    "get_gaia_DR2_catalog",
    "gen_tap_box_query",
    "query_tap_catalog",
    "get_hubble_source_catalog",
    "get_nsc_catalog",
    "get_desdr1_catalog",
    "get_skymapper_catalog",
    "get_vexas_catalog",
    "get_vhs_catalog",
    "get_panstarrs_catalog",
    "get_radec_catalog",
]


def table_to_radec(table, output="coords.radec"):
    """
    Make a "radec" ascii file with ra, dec columns from a table object

    Parameters
    ----------
    table : `~astropy.table.Table`
        Table object

    output : str
        Output filename. Default is `coords.radec`.

    """

    if "X_WORLD" in table.colnames:
        rc, dc = "X_WORLD", "Y_WORLD"
    elif "x_world" in table.colnames:
        rc, dc = "x_world", "y_world"
    elif "ra" in table.colnames:
        rc, dc = "ra", "dec"
    else:
        raise ValueError("Couldn't identify sky coordinates (x_world, ra)")

    table[rc].format = ".8f"
    table[dc].format = ".8f"

    table[rc, dc].write(output, format="ascii.commented_header", overwrite=True)


def table_to_regions(
    table,
    output="ds9.reg",
    comment=None,
    header="global color=green width=1",
    size=0.5,
    use_ellipse=False,
    use_world=True,
    verbose=True,
    unit_offset_colums=["x"],
    scale_major=1,
    extra=None,
):
    """
    Make a DS9 region file from a table object
    
    Parameters
    ----------

    table : `~astropy.table.Table`
        Table object

    output : str
        Output filename. Default is `ds9.reg`.

    comment : list of str
        Comment for each region. Default is None.

    header : str 
        DS9 region file header. Default `global color=green width=1`.
    
    size : float
        Default size for circles. Default is 0.5 arcsec.
    
    use_ellipse : bool
        Use ellipse regions if available. Default is False.

    use_world : bool
        Use "WORLD" coordinates if available. Default is True.

    verbose : bool
        Messaging to terminal. Default is True.

    unit_offset_colums : list
        Columns that are 1-indexed. Default is ["x"].

    scale_major : float
        Scale factor for major axis of ellipse regions. Default is 1.

    extra : list of str
        Extra text to add to each region line. Default is None.

    """
    fp = open(output, "w")
    fp.write(header + "\n")

    if ("X_WORLD" in table.colnames) & (use_world):
        xc, yc = "X_WORLD", "Y_WORLD"
        is_world = True
    elif ("ra" in table.colnames) & (use_world):
        xc, yc = "ra", "dec"
        is_world = True
    elif ("x_world" in table.colnames) & (use_world):
        xc, yc = "x_world", "y_world"
        is_world = True
    elif "x_image" in table.colnames:
        xc, yc = "x_image", "y_image"
        is_world = False
    elif "x" in table.colnames:
        xc, yc = "x", "y"
        is_world = False
    else:
        raise ValueError("Couldn't identify ra/dec or x/y columns")

    xdata = table[xc] * 1
    ydata = table[yc] * 1

    if xc in unit_offset_colums:
        if verbose:
            print(f"Unit offset for columns {xc}, {yc}")

        xdata += 1
        ydata += 1

    if is_world:
        fp.write("icrs\n")
        sec = '"'
    else:
        fp.write("image\n")
        sec = ""

    # GAIA
    if "solution_id" in table.colnames:
        e = np.sqrt(table["ra_error"] ** 2 + table["dec_error"] ** 2) / 1000.0
        e = np.maximum(e, 0.1)
    else:
        e = np.ones(len(table)) * size

    if use_ellipse:
        if is_world:
            if "a_world" in table.colnames:
                amaj = table["a_world"] * 3600 * scale_major
                amin = table["b_world"] * 3600 * scale_major
                etheta = table["theta_world"] / np.pi * 180  # +90
            else:
                use_ellipse = False
        else:
            if "a_image" in table.colnames:
                amaj = table["a_image"] * scale_major
                amin = table["b_image"] * scale_major
                etheta = table["theta_image"] / np.pi * 180  # +90
            else:
                use_ellipse = False

    if verbose:
        print(f"{output}: x = {xc}, y={yc}, ellipse={use_ellipse}")

    if use_ellipse:
        regstr = "ellipse({x:.7f}, {y:.7f}, {a:.3f}{sec}, {b:.3f}{sec}, {theta:.1f})"
        lines = [
            regstr.format(
                x=xdata[i], y=ydata[i], a=amaj[i], b=amin[i], sec=sec, theta=etheta[i]
            )
            for i in range(len(table))
        ]

    else:
        regstr = "circle({0:.7f}, {1:.7f}, {2:.3f}{3})"
        lines = [
            regstr.format(xdata[i], ydata[i], e[i], sec) for i in range(len(table))
        ]

    if comment is not None:
        for i in range(len(table)):
            lines[i] += " # text={{{0}}}".format(comment[i])

    if extra is not None:
        for i, li in enumerate(lines):
            if "#" not in li:
                lines[i] += " #"

            lines[i] += " " + extra[i]

    # newline
    lines = [l + "\n" for l in lines]

    fp.writelines(lines)
    fp.close()


def randomize_segmentation_labels(seg, random_seed=1, fill_value=np.nan):
    """
    Randomize labels on a segmentation image for easier visualization

    Parameters
    ----------
    seg : str or array
        Segmentation filename or integer array

    random_seed : int
        Random number seed for `numpy.random.seed`

    fill_value : scalar
        Value to insert where ``seg == 0``

    Returns
    -------
    rand_ids : array
        1D array of the randomized labels with size `max(seg)+1`

    rand_seg : array
        2D rray with randomized labels

    """

    np.random.seed(random_seed)

    if isinstance(seg, str):
        seg_im = pyfits.open(seg)
        seg_data = seg_im[0].data
    else:
        seg_data = seg

    imax = seg_data.max()

    rand_ids = np.append(fill_value, np.argsort(np.random.rand(imax)) + 1)
    return rand_ids, rand_ids[seg_data]


def get_ukidss_catalog(
    ra=165.0, dec=34.8, radius=3, database="UKIDSSDR9PLUS", programme_id="LAS"
):
    """
    Query for objects in the UKIDSS catalogs

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees

    radius : float
        Radius of the query, in arcmin

    database : str
        UKIDSS database to query.  Default is `UKIDSSDR9PLUS`.

    programme_id : str
        UKIDSS programme ID.  Default is `LAS`.

    Returns
    -------
    table : `~astropy.table.Table`
        Result of the query

    """

    from astroquery.ukidss import Ukidss

    coo = coord.SkyCoord(ra * u.deg, dec * u.deg)

    table = Ukidss.query_region(
        coo, radius=radius * u.arcmin, database=database, programme_id=programme_id
    )

    return table


def get_sdss_catalog(ra=165.86, dec=34.829694, radius=3):
    """
    Query for objects in the SDSS photometric catalog

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees

    radius : float
        Radius of the query, in arcmin

    Returns
    -------
    table : `~astropy.table.Table`
        Result of the query

    """
    from astroquery.sdss import SDSS

    coo = coord.SkyCoord(ra * u.deg, dec * u.deg)

    fields = ["ra", "dec", "raErr", "decErr", "petroMag_r", "petroMagErr_r"]
    # print fields
    fields = None

    table = SDSS.query_region(
        coo, radius=radius * u.arcmin, spectro=False, photoobj_fields=fields
    )

    return table


def get_twomass_catalog(ra=165.86, dec=34.829694, radius=3, catalog="fp_psc"):
    """
    Query for objects in the IRSA 2MASS catalog

    Wrapper for `astroquery.irsa.Irsa.query_region`

    Parameters
    ----------

    ra, dec : float
        Center of the query region, decimal degrees

    radius : float
        Radius of the query, in arcmin

    """
    return get_irsa_catalog(
        ra=ra, dec=dec, radius=radius, catalog=catalog, wise=False, twomass=True
    )


def get_irsa_catalog(
    ra=165.86,
    dec=34.829694,
    radius=3,
    catalog="allwise_p3as_psd",
    wise=False,
    twomass=False,
    ROW_LIMIT=500000,
    TIMEOUT=3600,
):
    """
    Query for objects in the `AllWISE <http://wise2.ipac.caltech.edu/docs/release/allwise/>`_ source catalog

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees

    radius : float
        Radius of the query, in arcmin

    catalog : str
        Catalog to query.  Default is `allwise_p3as_psd`.  2MASS is `fp_psc`.

    wise : bool
        Query the WISE catalog.  Default is False.
    
    twomass : bool
        Query the 2MASS catalog.  Default is False.

    ROW_LIMIT : int
        Maximum number of rows to return.  Default is 500000.

    TIMEOUT : int
        Timeout for the query.  Default is 3600.

    Returns
    -------
    table : `~astropy.table.Table`
        Result of the query

    """
    from astroquery.irsa import Irsa

    Irsa.ROW_LIMIT = ROW_LIMIT
    Irsa.TIMEOUT = TIMEOUT

    # all_wise = 'wise_allwise_p3as_psd'
    # all_wise = 'allwise_p3as_psd'
    if wise:
        catalog = "allwise_p3as_psd"
    elif twomass:
        catalog = "fp_psc"

    coo = coord.SkyCoord(ra * u.deg, dec * u.deg)

    table = Irsa.query_region(
        coo,
        catalog=catalog,
        spatial="Cone",
        radius=radius * u.arcmin,
        get_query_payload=False,
    )

    return table


def get_gaia_radec_at_time(gaia_tbl, date=2015.5, format="decimalyear"):
    """
    Use `~astropy.coordinates.SkyCoord.apply_space_motion` to compute 
    GAIA positions at a specific observation date

    Parameters
    ----------
    gaia_tbl : `~astropy.table.Table`
        GAIA table query, e.g., provided by `get_gaia_DR2_catalog`.

    date : e.g., float
        Observation date that can be parsed with `~astropy.time.Time`

    format : str
        Date format, see `~astropy.time.Time.FORMATS`.

    Returns
    -------
    coords : `~astropy.coordinates.SkyCoord`
        Projected sky coordinates.

    """
    from astropy.time import Time
    from astropy.coordinates import SkyCoord

    # Distance and radial_velocity are dummy numbers needed
    # to get the space motion correct

    try:
        # Try to use pyia
        import pyia

        g = pyia.GaiaData(gaia_tbl)
        coord = g.get_skycoord(
            distance=1 * u.kpc, frame="icrs", radial_velocity=0.0 * u.km / u.second
        )
    except:
        # From table itself
        if "ref_epoch" in gaia_tbl.colnames:
            ref_epoch = Time(gaia_tbl["ref_epoch"].data, format="decimalyear")
        else:
            ref_epoch = Time(2015.5, format="decimalyear")

        pm_de_column = "pmde" if "pmde" in gaia_tbl.colnames else "pmdec"

        coord = SkyCoord(
            ra=gaia_tbl["ra"],
            dec=gaia_tbl["dec"],
            pm_ra_cosdec=gaia_tbl["pmra"],
            pm_dec=gaia_tbl[pm_de_column],
            obstime=ref_epoch,
            frame="icrs",
            distance=1 * u.kpc,
            radial_velocity=0.0 * u.km / u.second,
        )
    
    if date > 50000:
        format = 'mjd'

    new_obstime = Time(date, format=format)
    coord_at_time = coord.apply_space_motion(new_obstime=new_obstime)
    return coord_at_time


# GAIA_DR2_COLUMNS = None
GAIA_DR2_COLUMNS = [
    "RA_ICRS",
    "DE_ICRS",
    "Epoch",
    "e_RA_ICRS",
    "e_DE_ICRS",
    "pmRA",
    "e_pmRA",
    "pmDE",
    "e_pmDE",
    # "NAL",
    # "NAC",
    # "Solved",
    # "APF",
    # "WAL",
    "RPmag",
]


def get_gaia_DR2_vizier_columns():
    """
    Get translation of Vizier GAIA DR2 columns.  The subset of desired
    output columns can be specified in the GAIA_DR2_COLUMNS global variable,
    which, if `None`, will be all available columns.
    """

    file = os.path.join(os.path.dirname(__file__), "data/gaia_dr2_vizier_columns.txt")
    lines = open(file).readlines()[1:]

    gdict = OrderedDict()
    for line in lines:
        viz_col = line.split()[0]
        gaia_col = line.split()[1][1:-1]
        if GAIA_DR2_COLUMNS is not None:
            if viz_col in GAIA_DR2_COLUMNS:
                gdict[viz_col] = gaia_col
        else:
            gdict[viz_col] = gaia_col

    return gdict


def get_gaia_vizier(
    ra=165.86,
    dec=34.829694,
    radius=3.0,
    max=100000,
    catalog="I/355/gaiadr3",
    keys=None,
    extra_query=" AND RPmag < 21",
    mjd=None,
    clean_mjd=True,
    lowercase=True,
    **kwargs,
):
    """
    Query GAIA catalogs from Vizier TAP

    Parameters
    ----------

    ra, dec : float
        Center of the query region, decimal degrees

    radius : float
        Radius of the query, in arcmin

    max : int
        Maximum number of rows to return.  Default is 100000.

    catalog : str
        Vizier catalog name.  Default is `I/345/gaia2`.
            DR2: ``catalog="I/345/gaia2"``
            EDR3: ``catalog="I/350/gaiaedr3"``
            DR3: ``catalog="I/355/gaiadr3"`

    keys : list of str
        List of column keys to query.  Default is None.

    extra_query : str
        Additional query criteria, e.g., magnitude limits

    mjd : float
        MJD to compute proper motions.  Default is None.

    clean_mjd : bool
        Remove rows with NaN proper motions.  Default is True.

    Returns
    -------

    table : `~astropy.table.Table`
        Result of the query

    """
    gdict = get_gaia_DR2_vizier_columns()
    if keys is None:
        keys = list(gdict.keys())

    if "dr3" in catalog:
        for k in ["Epoch","WAL"]:
            if k in keys:
                keys.pop(keys.index(k))

    try:
        result = query_tap_catalog(
            ra,
            dec,
            radius=radius,
            db=f"\"{catalog}\"",
            columns=[f'"{k}"' for k in keys],
            extra=extra_query,
            max=max,
            vizier=True,
            circle=True,
        )
    except:
        utils.log_exception(utils.LOGFILE, traceback)
        return False

    if ("dr3" in catalog) & (len(result) > 0):
        result["ref_epoch"] = 2016.0
    else:
        result["ref_epoch"] = 2015.5

    result["ra"] = result["RA_ICRS"]
    result["dec"] = result["DE_ICRS"]

    if lowercase:
        cols = result.colnames
        for c in cols:
            if c != c.lower():
                result.rename_column(c, c.lower())

    if mjd is not None:
        rd = get_gaia_radec_at_time(result, date=mjd, format="mjd")

        result["ra_time"] = rd.ra.deg
        result["dec_time"] = rd.dec.deg

        result.meta["cootime"] = (mjd, "Specified MJD for ra/dec_time")

        if clean_mjd:
            ok = np.isfinite(rd.ra.deg + rd.dec.deg)
            result = result[ok]

    return result


def get_gaia_DR2_vizier(
    ra=165.86,
    dec=34.829694,
    radius=3.0,
    max=100000,
    catalog="I/355/gaiadr3",
    server="vizier.u-strasbg.fr",
    keys=None,
    mjd=None,
    clean_mjd=True,
):
    """
    Query GAIA catalog from Vizier

    *Deprecated:* use `~grizli.catalog.get_gaia_vizier`

    Parameters
    ----------

    ra, dec : float
        Center of the query region, decimal degrees

    radius : float
        Radius of the query, in arcmin

    max : int
        Maximum number of rows to return.  Default is 100000.

    catalog : str
        Vizier catalog name.  Default is `I/345/gaia2`.
            DR2: ``catalog="I/345/gaia2"``
            EDR3: ``catalog="I/350/gaiaedr3"``
            DR3: ``catalog="I/355/gaiadr3"`

    server : str
        Vizier server.  Default is `vizier.u-strasbg.fr`.

    keys : list of str
        List of column keys to query.  Default is None.

    mjd : float
        MJD to compute proper motions.  Default is None.

    clean_mjd : bool
        Remove rows with NaN proper motions.  Default is True.

    Returns
    -------

    table : `~astropy.table.Table`
        Result of the query

    """
    from astroquery.vizier import Vizier
    import astropy.units as u
    import astropy.coordinates as coord

    coo = coord.SkyCoord(ra=ra, dec=dec, unit=(u.deg, u.deg), frame="icrs")

    gdict = get_gaia_DR2_vizier_columns()
    if keys is None:
        keys = list(gdict.keys())

    try:
        # Hack, Vizier object doesn't seem to allow getting all keys
        # simultaneously (astroquery v0.3.7)
        N = 9
        for i in range(len(keys) // N + 1):
            v = Vizier(catalog=catalog, columns=["+_r"] + keys[i * N : (i + 1) * N])
            v.VIZIER_SERVER = server
            v.ROW_LIMIT = max
            tab = v.query_region(coo, radius="{0}m".format(radius), catalog=catalog)[0]
            if i == 0:
                result = tab
            else:
                for k in tab.colnames:
                    # print(i, k)
                    result[k] = tab[k]

        for k in gdict:
            if k in result.colnames:
                result.rename_column(k, gdict[k])
    except:
        utils.log_exception(utils.LOGFILE, traceback)
        return False

    if mjd is not None:
        rd = get_gaia_radec_at_time(result, date=mjd, format="mjd")
        result["ra_time"] = rd.ra.deg
        result["dec_time"] = rd.dec.deg
        result.meta["cootime"] = mjd, "Specified MJD for ra/dec_time"
        if clean_mjd:
            ok = np.isfinite(rd.ra.deg + rd.dec.deg)
            result = result[ok]

    return result


def gaia_catalog_for_assoc(assoc_name="j191132p1652_hen-2-427-f335m_00007", min_radius=3, **kwargs):
    """
    Get GAIA catalog

    Parameters
    ----------
    assoc_name : str
        Name of the association. 
        Dedfault is `j191132p1652_hen-2-427-f335m_00007`.

    min_radius : float
        Minimum query radius, arcmin

    Returns
    -------
    gaia : `~astropy.table.Table`
        Result of the query.

    """
    from grizli.aws import db
    from grizli import utils

    res = db.SQL(
        f"""select t_min, t_max, filter, footprint
    from assoc_table where assoc_name = '{assoc_name}'
    """
    )

    sr = utils.SRegion(res["footprint"][0])
    for fp in res["footprint"][1:]:
        sri = utils.SRegion(fp)
        sr.xy.extend(sri.xy)

    un = sr.union()
    radius = np.maximum(np.sqrt(un.sky_area()[0]).value, min_radius)

    ra_center, de_center = un.centroid[0]

    gaia = get_gaia_vizier(ra_center, de_center, radius=radius, **kwargs)

    rd = get_gaia_radec_at_time(gaia, np.mean(res["t_min"]), format="mjd")

    gaia["ra_orig"] = gaia["ra"]
    gaia["de_orig"] = gaia["dec"]
    gaia["ra"] = rd.ra.deg
    gaia["dec"] = rd.dec.deg

    coo = np.array([gaia["ra"], gaia["dec"]])
    in_footprint = un.path[0].contains_points(coo.T)
    gaia["in_assoc"] = in_footprint

    gaia.meta["assocnam"] = assoc_name
    gaia["valid"] = np.isfinite(gaia["ra"] + gaia["dec"])

    # # Calculate PA relative to pointing center for potential spike contamination
    # dra = (gaia["ra_orig"] - ra_center)*np.cos(dec_center/180*np.pi)
    # dde = (gaia["de_orig"] - de_center)
    # gaia["pa"] = np.arctan2(-dra, dde)/np.pi*180

    return gaia


def gaia_dr2_conesearch_query(ra=165.86, dec=34.829694, radius=3.0, max=100000):
    """
    Generate a query string for the TAP servers

    Parameters
    ----------
    ra, dec : float
        RA, Dec in decimal degrees

    radius : float
        Search radius, in arc-minutes.

    max : int
        Maximum number of rows to return.  Default is 100000.

    Returns
    -------
    query : str
        Query string

    """
    query = "SELECT TOP {3} * FROM gaiadr2.gaia_source  WHERE CONTAINS(POINT('ICRS',gaiadr2.gaia_source.ra,gaiadr2.gaia_source.dec),CIRCLE('ICRS',{0},{1},{2:.2f}))=1".format(
        ra, dec, radius / 60.0, max
    )
    return query


def get_gaia_DR2_catalog(
    ra=165.86,
    dec=34.829694,
    radius=3.0,
    use_mirror=True,
    max_wait=20,
    max=100000,
    output_file="gaia.fits",
):
    """
    Query GAIA DR2 astrometric catalog

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees

    radius : float
        Radius of the query, in arcmin

    use_mirror : bool
        If True, use the mirror at `gaia.ari.uni-heidelberg.de`.  Otherwise
        use `gea.esac.esa.int`.

    max_wait : int
        Maximum number of seconds to wait for the query to complete.
        Default is 20.
    
    max : int
        Maximum number of rows to return.  Default is 100000.

    output_file : str
        Output filename.  Default is `gaia.fits`.

    Returns
    -------
    table : `~astropy.table.Table`
        Result of the query

    """
    try:
        import httplib
        from urllib import urlencode
    except:
        # python 3
        import http.client as httplib
        from urllib.parse import urlencode

    # import http.client in Python 3
    # import urllib.parse in Python 3
    import time
    from xml.dom.minidom import parseString

    host = "gea.esac.esa.int"
    port = 80
    pathinfo = "/tap-server/tap/async"

    if use_mirror:
        host = "gaia.ari.uni-heidelberg.de"
        pathinfo = "/tap/async"

    # -------------------------------------
    # Create job

    query = gaia_dr2_conesearch_query(
        ra=ra, dec=dec, radius=radius, max=max
    )  # "SELECT TOP 100000 * FROM gaiadr2.gaia_source  WHERE CONTAINS(POINT('ICRS',gaiadr2.gaia_source.ra,gaiadr2.gaia_source.dec),CIRCLE('ICRS',{0},{1},{2:.2f}))=1".format(ra, dec, radius/60.)
    print(query)

    params = urlencode(
        {
            "REQUEST": "doQuery",
            "LANG": "ADQL",
            "FORMAT": "fits",
            "PHASE": "RUN",
            "QUERY": query,
        }
    )

    headers = {
        "Content-type": "application/x-www-form-urlencoded",
        "Accept": "text/plain",
    }

    connection = httplib.HTTPConnection(host, port)
    connection.request("POST", pathinfo, params, headers)

    # Status
    response = connection.getresponse()
    print("Status: " + str(response.status), "Reason: " + str(response.reason))

    # Server job location (URL)
    location = response.getheader("location")
    print("Location: " + location)

    # Jobid
    jobid = location[location.rfind("/") + 1 :]
    print("Job id: " + jobid)

    connection.close()

    # -------------------------------------
    # Check job status, wait until finished

    tcount = 0
    while True:
        connection = httplib.HTTPConnection(host, port)
        connection.request("GET", pathinfo + "/" + jobid)
        response = connection.getresponse()
        data = response.read()
        # XML response: parse it to obtain the current status
        dom = parseString(data)

        if use_mirror:
            phaseElement = dom.getElementsByTagName("phase")[0]
        else:
            phaseElement = dom.getElementsByTagName("uws:phase")[0]

        phaseValueElement = phaseElement.firstChild
        phase = phaseValueElement.toxml()
        print("Status: " + phase)

        # Check finished
        if phase == "COMPLETED":
            break

        # wait and repeat
        time.sleep(0.2)
        tcount += 0.2

        if (phase == "ERROR") | (tcount > max_wait):
            return False

    # print "Data:"
    # print data

    connection.close()

    # -------------------------------------
    # Get results
    connection = httplib.HTTPConnection(host, port)
    connection.request("GET", pathinfo + "/" + jobid + "/results/result")
    response = connection.getresponse()
    data = response.read()
    outputFileName = output_file + (not use_mirror) * ".gz"
    try:
        outputFile = open(outputFileName, "w")
        outputFile.write(data)
    except:
        # Python 3
        outputFile = open(outputFileName, "wb")
        outputFile.write(data)

    outputFile.close()
    connection.close()
    print("Data saved in: " + outputFileName)

    if not use_mirror:
        # ESA archive returns gzipped
        try:
            os.remove(output_file)
        except:
            pass

        os.system("gunzip {output_file}.gz".format(output_file=output_file))

    table = Table.read(output_file, format="fits")
    return table


def gen_tap_box_query(
    ra=165.86,
    dec=34.829694,
    radius=3.0,
    corners=None,
    max=100000,
    db="ls_dr7.tractor_primary",
    columns=["*"],
    rd_colnames=["ra", "dec"],
    wcs_pad=0.5,
    circle=False,
):
    """
    Generate a query string for the NOAO Legacy Survey TAP server

    Parameters
    ----------
    ra, dec : float
        RA, Dec in decimal degrees

    radius : float
        Search radius, in arc-minutes.

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `~astropy.wcs.WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method

    max : int
        Maximum number of rows to return.  Default is 100000.

    db : str
        Database to query.  Default is `ls_dr7.tractor_primary`.
    
    columns : list of str
        List of columns to output.  Default ['*'] returns all columns.

    rd_colnames : str, str
        Column names in `db` corresponding to ra/dec (degrees).
        Default is `["ra", "dec"]`.

    wcs_pad : float
        Padding to add to WCS box queries.  Default is 0.5.

    circle : bool
        Use a circle and point spatial query rather than a box

    Returns
    -------
    query : str
        Query string

    """

    rmi = radius / 60 / 2
    cosd = np.cos(dec / 180 * np.pi)

    if max is not None:
        maxsel = "TOP {0}".format(max)
    else:
        maxsel = ""

    if corners is not None:
        if hasattr(corners, "calc_footprint"):
            foot = corners.calc_footprint()

            left = foot[:, 0].min()
            right = foot[:, 0].max()
            bottom = foot[:, 1].min()
            top = foot[:, 1].max()

            dx = right - left
            dy = top - bottom
            left -= wcs_pad * dx
            right += wcs_pad * dx
            bottom -= wcs_pad * dy
            top += wcs_pad * dy

        elif len(corners) != 4:
            msg = "corners needs 4 values (ra_min, ra_max, dec_min, dec_max)"
            raise ValueError(msg)
        else:
            left, right, bottom, top = corners
    else:
        left = ra - rmi / cosd
        right = ra + rmi / cosd
        bottom = dec - rmi
        top = dec + rmi

    fmt = dict(
        ra=ra,
        dec=dec,
        radius_deg=radius/60.,
        rc=rd_colnames[0],
        dc=rd_colnames[1],
        left=left,
        right=right,
        top=top,
        bottom=bottom,
        maxsel=maxsel,
        db=db,
        output_columns=", ".join(columns),
    )

    if circle:
        query = (
            "SELECT {maxsel} {output_columns} FROM {db}"
            + " WHERE 1=CONTAINS(POINT('ICRS',{db}.RAJ2000, {db}.DEJ2000),"
            +  "CIRCLE('ICRS', {ra}, {dec}, {radius_deg:.3f}))"
        )
    elif not np.isfinite(ra + dec):
        query = "SELECT {maxsel} {output_columns} FROM {db} "
    else:
        query = (
            "SELECT {maxsel} {output_columns} FROM {db} WHERE "
            + "{rc} > {left} AND {rc} < {right} AND "
            + "{dc} > {bottom} AND {dc} < {top} "
        )

    return query.format(**fmt)


def query_tap_catalog(
    ra=165.86,
    dec=34.829694,
    radius=3.0,
    corners=None,
    db="ls_dr9.tractor",
    columns=["*"],
    extra="",
    rd_colnames=["ra", "dec"],
    tap_url="https://datalab.noirlab.edu/tap",
    max=1000000,
    clean_xml=True,
    verbose=True,
    des=False,
    gaia=False,
    nsc=False,
    vizier=False,
    skymapper=False,
    circle='auto',
    hubble_source_catalog=False,
    tap_kwargs={},
    **kwargs,
):
    """Query NOAO Catalog holdings

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees.  If ``ra is None``, ignore the
        spatial query and query ``TOP {max} * from {db} {extra}``.

    radius : float
        Radius of the query, in arcmin

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method

    db : str
        Parent database (https://datalab.noirlab.edu/query.php).

    columns : list of str
        List of columns to output.  Default ['*'] returns all columns.

    extra : str
        String to add to the end of the positional box query, e.g.,
        'AND mag_auto_i > 16 AND mag_auto_i < 16.5'.

    rd_colnames : str, str
        Column names in `db` corresponding to ra/dec (degrees).

    tap_url : str
        TAP hostname

    des : bool
        Query `des_dr1.main` from NOAO.

    gaia : bool
        Query `gaiadr2.gaia_source` from http://gea.esac.esa.int.

    nsc : bool
        Query the NOAO Source Catalog (Nidever et al. 2018), `nsc_dr1.object`.

    vizier : bool
        Use the VizieR TAP server at  http://tapvizier.u-strasbg.fr/TAPVizieR/tap, see http://tapvizier.u-strasbg.fr/adql/about.html.

    circle : bool, 'auto'
        Use a circle query rather than a box, which seems to be faster on Vizier.  If
        'auto' and ``vizier=True``, will default to ``circle=True``

    hubble_source_catalog : bool
        Query the Hubble Source Catalog (v3).  If no 'NumImages' criteria is
        found in `extra`, then add an additional requirement:

            >>> extra += 'AND NumImages > 1'

    tap_kwargs : dict
        Additional keyword arguments to pass to `~astroquery.utils.tap.TapPlus`.

    Returns
    -------
    table : `~astropy.table.Table`
        Result of the query

    """
    from astroquery.utils.tap.core import TapPlus

    # DES DR1
    if des:
        if verbose:
            print("Query DES DR1 from NOAO")

        db = "des_dr1.main"
        tap_url = "https://datalab.noirlab.edu/tap"

    # NOAO source catalog, seems to have some junk
    if nsc:
        if verbose:
            print("Query NOAO source catalog")

        db = "nsc_dr1.object"
        tap_url = "https://datalab.noirlab.edu/tap"
        extra += " AND nsc_dr1.object.flags = 0"

    # GAIA DR2
    if gaia:
        if verbose:
            print("Query GAIA DR2 from ESA")

        db = "gaiadr2.gaia_source"
        tap_url = "http://gea.esac.esa.int/tap-server/tap"

    # VizieR TAP server
    if vizier:
        if verbose:
            print("Query {0} from VizieR TAP server".format(db))

        tap_url = "http://tapvizier.u-strasbg.fr/TAPVizieR/tap"
        rd_colnames = ["RAJ2000", "DEJ2000"]
        if circle in ['auto']:
            circle = True
    else:
        if circle in ['auto']:
            circle = False

    if skymapper:
        if verbose:
            print("Query {0} from VizieR TAP server".format(db))

        tap_url = "http://tapvizier.u-strasbg.fr/TAPVizieR/tap"
        rd_colnames = ["RAICRS", "DEICRS"]

    if hubble_source_catalog:
        if db is None:
            db = "dbo.SumPropMagAutoCat"
        elif "dbo" not in db:
            db = "dbo.SumPropMagAutoCat"

        tap_url = "http://vao.stsci.edu/HSCTAP/tapservice.aspx"
        rd_colnames = ["MatchRA", "MatchDec"]
        if "NumImages" not in extra:
            extra += "AND NumImages > 1"

    tap = TapPlus(url=tap_url, **tap_kwargs)

    if ra is not None:
        query = gen_tap_box_query(
            ra=ra,
            dec=dec,
            radius=radius,
            max=max,
            db=db,
            columns=columns,
            rd_colnames=rd_colnames,
            corners=corners,
            circle=circle,
        )
    else:
        query = f'SELECT TOP {max} * FROM {db} '

    job = tap.launch_job(query + extra, dump_to_file=True, verbose=verbose)
    try:
        table = job.get_results()
        if clean_xml:
            if hasattr(job, "outputFile"):
                jobFile = job.outputFile
            else:
                jobFile = job.get_output_file()

            os.remove(jobFile)

        # Provide ra/dec columns
        for c, cc in zip(rd_colnames, ["ra", "dec"]):
            if (c in table.colnames) & (cc not in table.colnames):
                table[cc] = table[c]

        table.meta["TAPURL"] = tap_url, "TAP URL"
        table.meta["TAPDB"] = db, "TAP database name"
        table.meta["TAPQUERY"] = query + extra, "TAP query"
        table.meta["RAQUERY"] = ra, "Query central RA"
        table.meta["DECQUERY"] = dec, "Query central Dec"
        table.meta["RQUERY"] = radius, "Query radius, arcmin"

        if hubble_source_catalog:
            for col in table.colnames:
                if table[col].dtype == "object":
                    print("Reformat column: {0}".format(col))
                    strcol = list(table[col])
                    table[col] = strcol

    except:
        if hasattr(job, "outputFile"):
            jobFile = job.outputFile
        else:
            jobFile = job.get_output_file()

        print("Query failed, check {0} for error messages".format(jobFile))
        table = None

    return table


# Limit Hubble Source Catalog query to brighter sources in limited bands
HSCv3_FILTER_LIMITS = {
    "W3_F160W": 23.5,
    "W3_F140W": 23.5,
    "W3_F125W": 23.5,
    "W3_F110W": 23.5,
    "W3_F098M": 23.5,
    "W3_F105W": 23.5,
    "A_F814W": 23.5,
    "W3_F814W": 23.5,
    "A_F606W": 23.5,
    "W3_F606W": 23.5,
    "A_F850LP": 23.5,
    "W3_F850LP": 23.5,
    "A_F775W": 23.5,
    "W3_F775W": 23.5,
}

HSCv3_COLUMNS = [
    "MatchRA",
    "MatchDec",
    "CI",
    "CI_Sigma",
    "KronRadius",
    "KronRadius_Sigma",
    "Extinction",
    "TargetName",
    "NumImages",
    "NumFilters",
    "NumVisits",
    "DSigma",
]


def get_hubble_source_catalog(
    ra=0.0,
    dec=0.0,
    radius=3,
    corners=None,
    max=int(1e7),
    extra=" AND NumImages > 0",
    kron_max=0.45,
    dsigma_max=100,
    clip_singles=10 * u.arcsec,
    verbose=True,
    columns=HSCv3_COLUMNS,
    filter_limits=HSCv3_FILTER_LIMITS,
):
    """
    Query NOAO Source Catalog, which is aligned to GAIA DR1.

    The default `extra` query returns well-detected sources in red bands.

    filter_limits : query on individual HSC filter magnitudes

    Wrapper for `query_tap_catalog`.

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees.
        Default is 0.0, 0.0.

    radius : float
        Radius of the query, in arcmin.
        Default is 3.

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method

    max : int 
        Maximum number of rows to return. Default is 100000.
    
    extra : str
        Additional query string. Default is " AND NumImages > 0".

    kron_max : float
        Maximum Kron radius. Default is 0.45.

    dsigma_max : float
        Maximum DSigma. Default is 100.

    clip_singles : float
        Maximum separation for singles to keep. Default is 10 arcsec.

    verbose : bool
        Print information. Default is True.

    columns : list of str
        Columns to query. Default is HSCv3_COLUMNS.

    filter_limits : dict
        Query on individual HSC filter magnitudes. Default is HSCv3_FILTER_LIMITS.

    Returns
    -------
    tab : `~astropy.table.Table`
        Result of the query.

    """
    import astropy.table

    msg = "Query NOAO Source Catalog ({ra:.5f},{dec:.5f},{radius:.1f}')"
    print(msg.format(ra=ra, dec=dec, radius=radius))

    if kron_max is not None:
        extra += " AND KronRadius < {0}".format(kron_max)
    if dsigma_max is not None:
        extra += " AND DSigma < {0}".format(dsigma_max)

    if filter_limits is not None:
        limit_list = ["{0} < {1}".format(f, filter_limits[f]) for f in filter_limits]

        filter_selection = " AND ({0})".format(" OR ".join(limit_list))
        extra += filter_selection

        columns += [f for f in filter_limits]

        db = "dbo.SumPropMagAutoCat p join dbo.SumMagAutoCat m on p.MatchID = m.MatchID"
    else:
        db = "dbo.SumPropMagAutoCat"

    tab = query_tap_catalog(
        ra=ra,
        dec=dec,
        radius=radius,
        corners=corners,
        max=max,
        extra=extra,
        hubble_source_catalog=True,
        verbose=verbose,
        db=db,
        columns=columns,
    )

    if clip_singles not in [None, False]:
        rr = tab["NumImages"] > 1
        if (rr.sum() > 0) & ((~rr).sum() > 0):
            r0, r1 = tab[rr], tab[~rr]
            idx, dr = utils.GTable(r0).match_to_catalog_sky(r1)
            new = dr > clip_singles
            xtab = astropy.table.vstack([r0, r1[new]])
            if verbose:
                msg = "HSCv3: Remove {0} NumImages == 1 sources " + "with tolerance {1}"
                print(msg.format((~new).sum(), clip_singles))

    return tab


def get_nsc_catalog(
    ra=0.0,
    dec=0.0,
    radius=3,
    corners=None,
    max=100000,
    extra=" AND (rerr < 0.08 OR ierr < 0.08 OR zerr < 0.08) AND raerr < 0.2 AND decerr < 0.2",
    verbose=True,
):
    """
    Query NOAO Source Catalog, which is aligned to GAIA DR1.

    The default `extra` query returns well-detected sources in red bands.

    Wrapper for `query_tap_catalog`.

    Parameters
    ----------

    ra, dec : float
        Center of the query region, decimal degrees.
        Default is 0.0, 0.0.

    radius : float
        Radius of the query, in arcmin.
        Default is 3.

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method

    max : int
        Maximum number of rows to return. Default is 100000.

    extra : str
        Additional query string.
        Default is " AND (rerr < 0.08 OR ierr < 0.08 OR zerr < 0.08) AND raerr < 0.2 AND decerr < 0.2".

    verbose : bool
        Print information. Default is True.

    Returns
    -------
    tab : `~astropy.table.Table`
        Result of the query.

    """
    msg = "Query NOAO Source Catalog ({ra:.5f},{dec:.5f},{radius:.1f}')"
    print(msg.format(ra=ra, dec=dec, radius=radius))

    tab = query_tap_catalog(
        ra=ra,
        dec=dec,
        radius=radius,
        corners=corners,
        extra=extra,
        nsc=True,
        verbose=verbose,
        max=max,
    )
    return tab


def get_desdr1_catalog(
    ra=0.0,
    dec=0.0,
    radius=3,
    corners=None,
    max=100000,
    extra=" AND (e_rmag < 0.15 OR e_imag < 0.15)",
    verbose=True,
):
    """
    Query DES DR1 Catalog from Vizier

    The default `extra` query returns well-detected sources in one or more
    red bands.

    Wrapper for `query_tap_catalog`.

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees.
        Default is 0.0, 0.0.

    radius : float
        Radius of the query, in arcmin.
        Default is 3.

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method

    max : int
        Maximum number of rows to return. Default is 100000.

    extra : str
        Additional query string.
        Default is " AND (e_rmag < 0.15 OR e_imag < 0.15)".

    verbose : bool
        Print information. Default is True.

    Returns
    -------
    tab : `~astropy.table.Table`
        Result of the query
    
    """
    msg = "Query DES Source Catalog ({ra:.5f},{dec:.5f},{radius:.1f}')"
    print(msg.format(ra=ra, dec=dec, radius=radius))

    tab = query_tap_catalog(
        ra=ra,
        dec=dec,
        radius=radius,
        corners=corners,
        extra=extra,
        db='"II/357/des_dr1"',
        vizier=True,
        verbose=verbose,
        max=max,
    )
    return tab


def get_desdr1_catalog_old(
    ra=0.0,
    dec=0.0,
    radius=3,
    corners=None,
    max=100000,
    extra=" AND (magerr_auto_r < 0.15 OR magerr_auto_i < 0.15)",
    verbose=True,
):
    """
    Query DES DR1 Catalog.

    The default `extra` query returns well-detected sources in one or more
    red bands.

    Wrapper for `query_tap_catalog`.

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees.
        Default is 0.0, 0.0.

    radius : float
        Radius of the query, in arcmin.
        Default is 3.

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method

    max : int
        Maximum number of rows to return. Default is 100000.

    extra : str
        Additional query string.
        Default is " AND (magerr_auto_r < 0.15 OR magerr_auto_i < 0.15)".

    verbose : bool
        Print information. Default is True.

    Returns
    -------
    tab : `~astropy.table.Table`
        Result of the query.

    """
    msg = "Query DES Source Catalog ({ra:.5f},{dec:.5f},{radius:.1f}')"
    print(msg.format(ra=ra, dec=dec, radius=radius))

    tab = query_tap_catalog(
        ra=ra,
        dec=dec,
        radius=radius,
        corners=corners,
        extra=extra,
        des=True,
        verbose=verbose,
        max=max,
    )
    return tab


def get_vhs_catalog(
    ra=0.0,
    dec=0.0,
    radius=3.0,
    corners=None,
    max_records=500000,
    verbose=True,
    extra="",
    table="vhs_dr4",
):
    """
    VISTA Hemisphere Survey (VHS) DR4.1 from vizier

    https://vizier.cds.unistra.fr/viz-bin/VizieR?-source=II/369

    Wrapper for `query_tap_catalog`.

    Parameters
    ----------

    ra, dec : float
        Center of the query region, decimal degrees.
        Default is 0.0, 0.0.

    radius : float
        Radius of the query, in arcmin.
        Default is 3.0.

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method

    max_records : int
        Maximum number of rows to return. Default is 500000.

    verbose : bool
        Print information. Default is True.

    extra : str
        Additional query string. Default is "".

    table : str
        Table to query. Default is "vhs_dr4".

    Returns
    -------

    tab : `~astropy.table.Table`
        Result of the query.

    """
    msg = "Query VHS DR4 {table} catalog ({ra},{dec},{radius})"
    print(msg.format(table=table, ra=ra, dec=dec, radius=radius))
    tab = query_tap_catalog(
        ra=ra,
        dec=dec,
        radius=radius * 2,
        corners=corners,
        extra=extra,
        vizier=True,
        db=f'"II/359/{table}"',
        verbose=verbose,
        max=max_records,
    )
    return tab


def get_vexas_catalog(
    ra=0.0,
    dec=0.0,
    radius=3.0,
    corners=None,
    max_records=500000,
    verbose=True,
    extra="",
    table="vexasdes",
):
    """
    VEXAS DR2 from vizier

    https://vizier.cds.unistra.fr/viz-bin/VizieR?-source=II/369

    Wrapper for `query_tap_catalog`.

    Parameters
    ----------

    ra, dec : float
        Center of the query region, decimal degrees.
        Default is 0.0, 0.0.

    radius : float
        Radius of the query, in arcmin.
        Default is 3.0.

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method


    max_records : int
        Maximum number of rows to return. Default is 500000.

    verbose : bool
        Print information. Default is True.

    extra : str
        Additional query string. Default is "".

    table : str
        Table to query. Default is "vexasdes".

    Returns
    -------
    tab : `~astropy.table.Table`
        Result of the query.

    """
    msg = "Query VEXAS DR2 {table} catalog ({ra},{dec},{radius})"
    print(msg.format(table=table, ra=ra, dec=dec, radius=radius))
    tab = query_tap_catalog(
        ra=ra,
        dec=dec,
        radius=radius * 2,
        corners=corners,
        extra=extra,
        vizier=True,
        db=f'"II/369/{table}"',
        verbose=verbose,
        max=max_records,
    )
    return tab


def get_skymapper_catalog(
    ra=0.0,
    dec=0.0,
    radius=3.0,
    corners=None,
    max_records=500000,
    verbose=True,
    extra="",
):
    """
    Get Skymapper DR1 from Vizier

    Wrapper for `query_tap_catalog`.

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees.
        Default is 0.0, 0.0.

    radius : float
        Radius of the query, in arcmin.
        Default is 3.0.

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method

    max_records : int
        Maximum number of rows to return. Default is 500000.

    verbose : bool
        Print information. Default is True.

    extra : str
        Additional query string. Default is "".

    Returns
    -------
    tab : `~astropy.table.Table`
        Result of the query.

    """
    msg = "Query Skymapper DR1 catalog ({ra},{dec},{radius})"
    print(msg.format(ra=ra, dec=dec, radius=radius))
    tab = query_tap_catalog(
        ra=ra,
        dec=dec,
        radius=radius * 2,
        corners=corners,
        extra=extra,
        skymapper=True,
        db='"II/358/smss"',
        verbose=verbose,
        max=max_records,
    )
    return tab


def get_legacysurveys_catalog(
    ra=0.0,
    dec=0.0,
    radius=3.0,
    verbose=True,
    db="ls_dr9.tractor",
    sn_lim=("r", 10),
    **kwargs,
):
    """
    Query LegacySurveys TAP catalog

    Wrapper for `query_tap_catalog`.

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees
    
    radius : float
        Radius of the query, in arcmin

    verbose : bool
        Print information. Default is True.

    db : str
        Database to query. Default is "ls_dr9.tractor".

    sn_lim : tuple
        Signal-to-noise limit. Default is ("r", 10).
    
    kwargs : dict
        Additional keyword arguments to pass to `query_tap_catalog`.

    Returns
    -------
    tab : `~astropy.table.Table`
        Result of the query

    """
    if verbose:
        msg = "Query LegacySurveys ({db}) catalog ({ra},{dec},{radius:.2f})"
        print(msg.format(ra=ra, dec=dec, radius=radius, db=db))

    if sn_lim is not None:
        if len(sn_lim) >= 2:
            b = sn_lim[0]
            extra = f" AND flux_{b}*SQRT(flux_ivar_{b}) > {sn_lim[1]}"
            if "extra" in kwargs:
                kwargs["extra"] += extra
            else:
                kwargs["extra"] = extra

    tab = query_tap_catalog(ra=ra, dec=dec, radius=radius, db=db, **kwargs)
    # if (sn_lim is not None) & (len(tab) > 1):
    #     if len(sn_lim) >= 2:
    #         sn = tab[f'flux_{sn_lim[0]}']
    #         sn *= np.sqrt(tab[f'flux_ivar_{sn_lim[0]}'])
    #
    #         valid = sn > sn_lim[1]
    #         return tab[valid]

    return tab


def get_panstarrs_catalog(
    ra=0.0,
    dec=0.0,
    radius=3.0,
    corners=None,
    max_records=500000,
    verbose=True,
    extra='AND "II/349/ps1".e_imag < 0.2 AND "II/349/ps1".e_RAJ2000 < 0.15 AND "II/349/ps1".e_DEJ2000 < 0.15',
):
    """
    Get PS1 from Vizier

    Wrapper for `query_tap_catalog`.

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees.
        Default is 0.0, 0.0.

    radius : float
        Radius of the query, in arcmin.
        Default is 3.0.

    corners : 4-tuple, `~astropy.wcs.WCS` or None
        ra_min, ra_max, dec_min, dec_max of a query box to use instead of
        `radius`.  Or if a `WCS` object, get limits from the
        `~astropy.wcs.WCS.calc_footprint` method

    max_records : int
        Maximum number of rows to return. Default is 500000.

    verbose : bool
        Print information. Default is True.

    extra : str
        Additional query string. Default is 'AND "II/349/ps1".e_imag < 0.2 AND "II/349/ps1".e_RAJ2000 < 0.15 AND "II/349/ps1".e_DEJ2000 < 0.15'.

    Returns
    -------
    tab : `~astropy.table.Table`
        Result of the query.

    """
    msg = "Query PanSTARRS catalog ({ra},{dec},{radius})"
    print(msg.format(ra=ra, dec=dec, radius=radius))
    tab = query_tap_catalog(
        ra=ra,
        dec=dec,
        radius=radius * 2,
        corners=corners,
        extra=extra,
        vizier=True,
        db='"II/349/ps1"',
        verbose=verbose,
        max=max_records,
    )
    return tab


def get_radec_catalog(
    ra=0.0,
    dec=0.0,
    radius=3.0,
    product="cat",
    verbose=True,
    reference_catalogs=[
        "GAIA",
        "LS_DR9",
        "PS1",
        "Hubble",
        "NSC",
        "SDSS",
        "WISE",
        "DES",
    ],
    use_self_catalog=False,
    **kwargs,
):
    """
    Decide what reference astrometric catalog to use

    First search SDSS, then WISE looking for nearby matches.

    Parameters
    ----------
    ra, dec : float
        Center of the query region, decimal degrees

    radius : float
        Radius of the query, in arcmin

    product : str
        Basename of the drizzled product. If a locally-created catalog with
        filename that startswith `product` is found, use that one instead of
        the external (low precision) catalogs so that you're matching
        HST-to-HST astrometry.

    verbose : bool
        Print information. Default is True.

    reference_catalogs : list
        Order in which to query reference catalogs.  Options are 'GAIA',
        'PS1' (STScI PanSTARRS), 'SDSS', 'WISE', 'NSC' (NOAO Source Catalog),
        'DES' (Dark Energy Survey DR1), 'Hubble' (Hubble Source Catalog v3),
        'LS_DR9', 'LS_DR10' (LegacySurveys DR9, DR10).

    Returns
    -------
    radec : str
        Filename of the RA/Dec list derived from the parent catalog

    ref_catalog : str, {'SDSS', 'WISE', 'VISIT'}
        Provenance of the `radec` list.

    """
    query_kwargs = {
        "LS_DR9": {"db": "ls_dr9.tractor"},
        "LS_DR10": {"db": "ls_dr10.tractor"},
    }

    query_functions = {
        "SDSS": get_sdss_catalog,
        "GAIA_TAP": get_gaia_DR2_catalog,
        "PS1": get_panstarrs_catalog,
        "WISE": get_irsa_catalog,
        "2MASS": get_twomass_catalog,
        "GAIA_Vizier": get_gaia_vizier,
        # "GAIA": get_gaia_DR2_vizier,
        "GAIA": get_gaia_vizier,
        "NSC": get_nsc_catalog,
        "DES": get_desdr1_catalog,
        "Hubble": get_hubble_source_catalog,
        "Skymapper": get_skymapper_catalog,
        "VEXAS": get_vexas_catalog,
        "LS_DR9": get_legacysurveys_catalog,
        "LS_DR10": get_legacysurveys_catalog,
    }

    # Try queries
    has_catalog = False
    ref_catalog = "None"
    ref_cat = []

    for ref_src in reference_catalogs:
        try:
            if ref_src == "GAIA":
                try:
                    ref_cat = query_functions[ref_src](
                        ra=ra, dec=dec, radius=radius, use_mirror=False
                    )
                except:
                    try:
                        ref_cat = query_functions[ref_src](
                            ra=ra, dec=dec, radius=radius
                        )
                    except:
                        ref_cat = False

                # Try GAIA mirror at Heidelberg
                if ref_cat is False:
                    ref_cat = query_functions[ref_src](
                        ra=ra, dec=dec, radius=radius, use_mirror=True
                    )
            else:
                if ref_src in query_kwargs:
                    _kws = query_kwargs[ref_src]
                else:
                    _kws = {}

                ref_cat = query_functions[ref_src](
                    ra=ra, dec=dec, radius=radius, **_kws
                )
            # #
            # ref_cat = query_functions[ref_src](ra=ra, dec=dec,
            #                                    radius=radius)

            valid = np.isfinite(ref_cat["ra"] + ref_cat["dec"])
            ref_cat = ref_cat[valid]

            if len(ref_cat) < 2:
                raise ValueError

            table_to_regions(
                ref_cat, output="{0}_{1}.reg".format(product, ref_src.lower())
            )
            ref_cat["ra", "dec"].write(
                "{0}_{1}.radec".format(product, ref_src.lower()),
                format="ascii.commented_header",
                overwrite=True,
            )

            radec = "{0}_{1}.radec".format(product, ref_src.lower())
            ref_catalog = ref_src
            has_catalog = True
            if len(ref_cat) > 0:
                break
        except:
            print("{0} query failed".format(ref_src))
            has_catalog = False

    if (ref_src.startswith("GAIA")) & ("date" in kwargs) & has_catalog:
        if kwargs["date"] is not None:
            if "date_format" in kwargs:
                date_format = kwargs["date_format"]
            else:
                date_format = "mjd"

            gaia_tbl = ref_cat  # utils.GTable.gread('gaia.fits')
            coo = get_gaia_radec_at_time(
                gaia_tbl, date=kwargs["date"], format=date_format
            )

            coo_tbl = utils.GTable()
            coo_tbl["ra"] = coo.ra
            coo_tbl["dec"] = coo.dec

            ok = np.isfinite(coo_tbl["ra"]) & np.isfinite(coo_tbl["dec"])

            coo_tbl.meta["date"] = kwargs["date"]
            coo_tbl.meta["datefmt"] = date_format

            msg = "Apply observation ({0},{1}) to GAIA catalog"
            print(msg.format(kwargs["date"], date_format))

            table_to_regions(
                coo_tbl[ok], output="{0}_{1}.reg".format(product, ref_src.lower())
            )

            coo_tbl["ra", "dec"][ok].write(
                "{0}_{1}.radec".format(product, ref_src.lower()),
                format="ascii.commented_header",
                overwrite=True,
            )

    if not has_catalog:
        return False

    # WISP, check if a catalog already exists for a given rootname and use
    # that if so.
    cat_files = glob.glob("-f1".join(product.split("-f1")[:-1]) + "-f*.cat*")
    if (len(cat_files) > 0) & (use_self_catalog):
        ref_cat = utils.GTable.gread(cat_files[0])
        root = cat_files[0].split(".cat")[0]
        ref_cat["X_WORLD", "Y_WORLD"].write(
            "{0}.radec".format(root), format="ascii.commented_header", overwrite=True
        )

        radec = "{0}.radec".format(root)
        ref_catalog = "VISIT"

    if verbose:
        msg = "{0} - Reference RADEC: {1} [{2}] N={3}"
        print(msg.format(product, radec, ref_catalog, len(ref_cat)))

    return radec, ref_catalog



"""
Pipeline for flexible modeling and extration of slitless spectroscopy
"""
from .version import __version__

import os

# Will get ImportError: No module named disperse" if imported in the repo directory
# if os.path.exists('README.rst') & os.path.exists('LICENSE.txt'):
#     print("""
# Warning: `import grizli` will fail if the working directory is the place
# where the code repository was cloned and compiled!
# """)

# Test that GRIZLI system variable is set
if os.getenv('GRIZLI') is None:
    GRIZLI_PATH = os.path.join(os.path.dirname(__file__), 'data/')
else:
    GRIZLI_PATH = os.getenv('GRIZLI')
